#
# sambacc: a samba container configuration tool
# Copyright (C) 2021  John Mulligan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>
#

import contextlib
import threading
import time

import pytest

try:
    import sambacc.inotify_waiter
except ImportError:
    pytestmark = pytest.mark.skip


@contextlib.contextmanager
def background(bg_func):
    t = threading.Thread(target=bg_func)
    t.start()
    try:
        yield None
    finally:
        t.join()


def test_inotify(tmp_path):
    tfile = str(tmp_path / "foobar.txt")
    tfile2 = str(tmp_path / "other.txt")

    iw = sambacc.inotify_waiter.INotify(tfile, print_func=print, timeout=3)

    def _touch():
        time.sleep(0.2)
        with open(tfile, "w") as fh:
            print("W", tfile)
            fh.write("one")

    with background(_touch):
        before = time.time()
        iw.wait()
        after = time.time()
    assert after - before > 0.1
    assert after - before <= 1

    def _touch2():
        time.sleep(0.2)
        with open(tfile2, "w") as fh:
            print("W", tfile2)
            fh.write("two")
        time.sleep(1)
        with open(tfile, "w") as fh:
            print("W", tfile)
            fh.write("one")

    with background(_touch2):
        before = time.time()
        iw.wait()
        after = time.time()

    assert after - before > 0.1
    assert after - before >= 1

    before = time.time()
    iw.wait()
    after = time.time()
    assert int(after) - int(before) == 3
    iw.close()


def test_inotify_bad_input():
    with pytest.raises(ValueError):
        sambacc.inotify_waiter.INotify("/")


def test_inotify_relative_path():
    iw = sambacc.inotify_waiter.INotify("cool.txt")
    assert iw._dir == "."
    assert iw._name == "cool.txt"



#
# sambacc: a samba container configuration tool
# Copyright (C) 2021  John Mulligan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>
#

from collections import namedtuple
import argparse
import logging
import typing

from sambacc import config
from sambacc import leader
from sambacc import permissions
from sambacc import simple_waiter

_INOTIFY_OK = True
try:
    from sambacc import inotify_waiter as iw
except ImportError:
    _INOTIFY_OK = False

_logger = logging.getLogger(__name__)


class Fail(ValueError):
    pass


class Parser(typing.Protocol):
    def frank(self, x: str) -> str:
        ...  # pragma: no cover

    def set_defaults(self, **kwargs: typing.Any) -> None:
        ...  # pragma: no cover

    def add_argument(
        self, *args: typing.Any, **kwargs: typing.Any
    ) -> typing.Any:
        ...  # pragma: no cover


Command = namedtuple("Command", "name cmd_func arg_func cmd_help")


def toggle_option(parser: Parser, arg: str, dest: str, helpfmt: str) -> Parser:
    parser.add_argument(
        arg,
        action="store_true",
        dest=dest,
        help=helpfmt.format("Enable"),
    )
    negarg = arg.replace("--", "--no-")
    parser.add_argument(
        negarg,
        action="store_false",
        dest=dest,
        help=helpfmt.format("Disable"),
    )
    return parser


def get_help(cmd: Command) -> str:
    if cmd.cmd_help is not None:
        return cmd.cmd_help
    if cmd.cmd_func.__doc__:
        return cmd.cmd_func.__doc__
    return ""


def add_command(subparsers: typing.Any, cmd: Command) -> None:
    subparser = subparsers.add_parser(cmd.name, help=get_help(cmd))
    subparser.set_defaults(cfunc=cmd.cmd_func)
    if cmd.arg_func is not None:
        cmd.arg_func(subparser)


class CommandBuilder:
    def __init__(self):
        self._commands = []
        self._names = set()

    def command(self, name, arg_func=None, cmd_help=None):
        if name in self._names:
            raise ValueError(f"{name} already in use")
        self._names.add(name)

        def _wrapper(f):
            self._commands.append(
                Command(
                    name=name, cmd_func=f, arg_func=arg_func, cmd_help=cmd_help
                )
            )
            return f

        return _wrapper

    def assemble(
        self, arg_func: typing.Optional[typing.Callable] = None
    ) -> argparse.ArgumentParser:
        parser = argparse.ArgumentParser()
        if arg_func is not None:
            arg_func(parser)
        subparsers = parser.add_subparsers()
        for cmd in self._commands:
            add_command(subparsers, cmd)
        return parser

    def dict(self) -> dict[str, Command]:
        """Return a dict mapping command names to Command object."""
        return {c.name: c for c in self._commands}


class Context(typing.Protocol):
    """Protocol type for CLI Context.
    Used to share simple, common state, derived from the CLI, across individual
    command functions.
    """

    # The expects_ctdb attribute indicates that the command can, and should,
    # make use of ctdb whenever ctdb is enabled in the configuration.
    expects_ctdb: bool

    @property
    def cli(self) -> argparse.Namespace:
        ...  # pragma: no cover

    @property
    def instance_config(self) -> config.InstanceConfig:
        ...  # pragma: no cover

    @property
    def require_validation(self) -> typing.Optional[bool]:
        ...  # pragma: no cover


def best_waiter(
    filename: typing.Optional[str] = None,
    max_timeout: typing.Optional[int] = None,
) -> simple_waiter.Waiter:
    """Fetch the best waiter type for our sambacc command."""
    if filename and _INOTIFY_OK:
        _logger.info("enabling inotify support")
        return iw.INotify(
            filename, print_func=_logger.info, timeout=max_timeout
        )
    # should max_timeout change Sleeper too? probably.
    return simple_waiter.Sleeper()


def best_leader_locator(
    iconfig: config.InstanceConfig,
) -> leader.LeaderLocator:
    """Fetch the best leader locator for our sambacc command.
    This only makes sense to be used in a clustered scenario.
    """
    from sambacc import ctdb

    return ctdb.CLILeaderLocator()


def perms_handler(
    config: config.PermissionsConfig,
    path: str,
) -> permissions.PermissionsHandler:
    """Fetch and instantiate the appropriate permissions handler for the given
    configuration.
    """
    if config.method == "none":
        _logger.info("Using no-op permissions handler")
        return permissions.NoopPermsHandler(
            path, config.status_xattr, options=config.options
        )
    if config.method == "initialize-share-perms":
        _logger.info("Using initializing posix permissions handler")
        return permissions.InitPosixPermsHandler(
            path, config.status_xattr, options=config.options
        )
    if config.method == "always-share-perms":
        _logger.info("Using always-setting posix permissions handler")
        return permissions.AlwaysPosixPermsHandler(
            path, config.status_xattr, options=config.options
        )
    # fall back to init perms handler
    _logger.info("Using initializing posix permissions handler")
    return permissions.InitPosixPermsHandler(
        path, config.status_xattr, options=config.options
    )


commands = CommandBuilder()
setup_steps = CommandBuilder()



#
# sambacc: a samba container configuration tool
# Copyright (C) 2021  John Mulligan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>
#

import os
import typing


class TextFileLoader:
    def __init__(self, path: str):
        self.path = path

    def read(self) -> None:
        with open(self.path) as f:
            self.readfp(f)

    def write(self) -> None:
        tpath = self._tmp_path(self.path)
        with open(tpath, "w") as f:
            self.writefp(f)
        os.rename(tpath, self.path)

    def _tmp_path(self, path: str) -> str:
        # for later: make this smarter
        return f"{path}.tmp"

    def readfp(self, fp: typing.IO) -> None:
        self.loadlines(fp.readlines())

    def writefp(self, fp: typing.IO) -> None:
        for line in self.dumplines():
            fp.write(line)
        fp.flush()

    def dumplines(self) -> typing.Iterable[str]:
        """Must be overridden."""
        return []

    def loadlines(self, lines: typing.Iterable[str]) -> None:
        """Must be overridden."""
        pass



#
# sambacc: a samba container configuration tool
# Copyright (C) 2021  John Mulligan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>
#

import json
import os
import pytest

from sambacc import samba_cmds
import sambacc.join


@pytest.fixture(scope="function")
def testjoiner(tmp_path):
    data_path = tmp_path / "_samba"
    data_path.mkdir()

    fake_net = [
        "#!/bin/sh",
        f'echo "ARGS:" "$@" > {data_path}/log',
        f"cat >> {data_path}/log",
        f"if grep -q failme {data_path}/log ; then exit 1; fi",
    ]
    fake_net_script = data_path / "net.sh"
    with open(fake_net_script, "w") as fh:
        fh.write("\n".join(fake_net))
        fh.write("\n")
    os.chmod(fake_net_script, 0o755)

    class TestJoiner(sambacc.join.Joiner):
        _net_ads_join = samba_cmds.SambaCommand(fake_net_script)["ads", "join"]
        path = tmp_path
        logpath = data_path / "log"

    return TestJoiner()


def test_no_sources(testjoiner):
    with pytest.raises(sambacc.join.JoinError):
        testjoiner.join()


def test_invalid_source_vals(testjoiner):
    with pytest.raises(ValueError):
        testjoiner.add_source("bob", 123)
    with pytest.raises(ValueError):
        testjoiner.add_source(sambacc.join.JoinBy.PASSWORD, 123)
    with pytest.raises(ValueError):
        testjoiner.add_source(sambacc.join.JoinBy.FILE, 123)


def test_join_password(testjoiner):
    testjoiner.add_source(
        sambacc.join.JoinBy.PASSWORD,
        sambacc.join.UserPass("bugs", "whatsupdoc"),
    )
    testjoiner.join()


def test_join_file(testjoiner):
    jpath1 = os.path.join(testjoiner.path, "join1.json")
    with open(jpath1, "w") as fh:
        json.dump({"username": "elmer", "password": "hunter2"}, fh)
    testjoiner.add_source(
        sambacc.join.JoinBy.FILE,
        jpath1,
    )
    testjoiner.join()


def test_join_missing_file(testjoiner):
    jpath1 = os.path.join(testjoiner.path, "nope.json")
    testjoiner.add_source(
        sambacc.join.JoinBy.FILE,
        jpath1,
    )
    with pytest.raises(sambacc.join.JoinError) as err:
        testjoiner.join()
    assert "not found" in str(err).lower()


def test_join_bad_file(testjoiner):
    jpath1 = os.path.join(testjoiner.path, "join1.json")
    testjoiner.add_source(
        sambacc.join.JoinBy.FILE,
        jpath1,
    )

    with open(jpath1, "w") as fh:
        json.dump({"acme": True}, fh)
    with pytest.raises(sambacc.join.JoinError):
        testjoiner.join()

    with open(jpath1, "w") as fh:
        json.dump({"username": None, "password": "hunter2"}, fh)
    with pytest.raises(sambacc.join.JoinError):
        testjoiner.join()

    with open(jpath1, "w") as fh:
        json.dump({"username": "elmer", "password": 123}, fh)
    with pytest.raises(sambacc.join.JoinError):
        testjoiner.join()


def test_join_multi_source(testjoiner):
    testjoiner.add_source(
        sambacc.join.JoinBy.PASSWORD,
        sambacc.join.UserPass("bugs", "whatsupdoc"),
    )
    jpath1 = os.path.join(testjoiner.path, "join1.json")
    with open(jpath1, "w") as fh:
        json.dump({"username": "elmer", "password": "hunter2"}, fh)
    testjoiner.add_source(
        sambacc.join.JoinBy.FILE,
        jpath1,
    )
    testjoiner.join()

    with open(testjoiner.logpath) as fh:
        lines = fh.readlines()
    assert lines[0].startswith("ARGS")
    assert "bugs" in lines[0]
    assert "whatsupdoc" in lines[1]


def test_join_multi_source_fail_first(testjoiner):
    testjoiner.add_source(
        sambacc.join.JoinBy.PASSWORD,
        sambacc.join.UserPass("bugs", "failme"),
    )
    jpath1 = os.path.join(testjoiner.path, "join1.json")
    with open(jpath1, "w") as fh:
        json.dump({"username": "elmer", "password": "hunter2"}, fh)
    testjoiner.add_source(
        sambacc.join.JoinBy.FILE,
        jpath1,
    )
    testjoiner.join()

    with open(testjoiner.logpath) as fh:
        lines = fh.readlines()
    assert lines[0].startswith("ARGS")
    assert "elmer" in lines[0]
    assert "hunter2" in lines[1]


def test_join_multi_source_fail_both(testjoiner):
    testjoiner.add_source(
        sambacc.join.JoinBy.PASSWORD,
        sambacc.join.UserPass("bugs", "failme"),
    )
    jpath1 = os.path.join(testjoiner.path, "join1.json")
    with open(jpath1, "w") as fh:
        json.dump({"username": "elmer", "password": "failme2"}, fh)
    testjoiner.add_source(
        sambacc.join.JoinBy.FILE,
        jpath1,
    )
    with pytest.raises(sambacc.join.JoinError) as err:
        testjoiner.join()
    assert err.match("2 join attempts")
    assert len(err.value.errors) == 2

    with open(testjoiner.logpath) as fh:
        lines = fh.readlines()
    assert lines[0].startswith("ARGS")
    assert "elmer" in lines[0]
    assert "failme2" in lines[1]


def test_join_prompt_fake(testjoiner):
    testjoiner.add_source(
        sambacc.join.JoinBy.INTERACTIVE,
        sambacc.join.UserPass("daffy"),
    )
    testjoiner.join()

    with open(testjoiner.logpath) as fh:
        lines = fh.readlines()

    assert lines[0].startswith("ARGS")
    assert "daffy" in lines[0]
    assert len(lines) == 1


def test_join_with_marker(testjoiner):
    testjoiner.marker = os.path.join(testjoiner.path, "marker.json")
    testjoiner.add_source(
        sambacc.join.JoinBy.PASSWORD,
        sambacc.join.UserPass("bugs", "whatsupdoc"),
    )
    testjoiner.join()

    assert os.path.exists(testjoiner.marker)
    assert testjoiner.did_join()


def test_join_bad_marker(testjoiner):
    testjoiner.marker = os.path.join(testjoiner.path, "marker.json")
    testjoiner.add_source(
        sambacc.join.JoinBy.PASSWORD,
        sambacc.join.UserPass("bugs", "whatsupdoc"),
    )
    testjoiner.join()

    # its ok after creation
    assert testjoiner.did_join()
    # invalid contents
    with open(testjoiner.marker, "w") as fh:
        json.dump({"foo": "bar"}, fh)
    assert not testjoiner.did_join()
    # missing file
    os.unlink(testjoiner.marker)
    assert not testjoiner.did_join()


def test_join_no_marker(testjoiner):
    testjoiner.add_source(
        sambacc.join.JoinBy.PASSWORD,
        sambacc.join.UserPass("bugs", "whatsupdoc"),
    )
    testjoiner.join()
    # join was successful, but no marker was set on the joiner
    # thus did_join must return false
    assert not testjoiner.did_join()


def test_join_when_possible(testjoiner):
    class DummyWaiter:
        wcount = 0

        def wait(self):
            self.wcount += 1

    waiter = DummyWaiter()

    errors = []

    def ehandler(err):
        nonlocal errors
        if len(errors) > 5:
            raise ValueError("xxx")
        errors.append(err)

    # error case - no valid join souces
    with pytest.raises(ValueError):
        sambacc.join.join_when_possible(
            testjoiner, waiter, error_handler=ehandler
        )

    assert len(errors) == 6
    assert waiter.wcount == 6

    # success case - performs a password join
    errors[:] = []
    testjoiner.add_source(
        sambacc.join.JoinBy.PASSWORD,
        sambacc.join.UserPass("bugs", "whatsupdoc"),
    )
    testjoiner.marker = os.path.join(testjoiner.path, "marker.json")
    sambacc.join.join_when_possible(testjoiner, waiter, error_handler=ehandler)

    assert len(errors) == 0
    assert waiter.wcount == 6
    with open(testjoiner.logpath) as fh:
        lines = fh.readlines()
    assert lines[0].startswith("ARGS")
    assert "bugs" in lines[0]
    assert "whatsupdoc" in lines[1]

    # success case - join marker exists
    sambacc.join.join_when_possible(testjoiner, waiter, error_handler=ehandler)

    assert len(errors) == 0
    assert waiter.wcount == 6



#
# sambacc: a samba container configuration tool
# Copyright (C) 2021  John Mulligan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>
#

import io
import json
import os

import pytest

import sambacc.config
import sambacc.samba_cmds
from sambacc import ctdb


def test_migrate_tdb(tmpdir, monkeypatch):
    src = tmpdir / "src"
    os.mkdir(src)
    dst = tmpdir / "dst"
    os.mkdir(dst)
    fake = tmpdir / "fake.sh"
    monkeypatch.setattr(ctdb, "_SRC_TDB_DIRS", [str(src)])
    monkeypatch.setattr(sambacc.samba_cmds, "_GLOBAL_PREFIX", [str(fake)])

    with open(fake, "w") as fh:
        fh.write("#!/bin/sh\n")
        fh.write('[ "$1" == "ltdbtool" ] || exit 1\n')
        fh.write('[ "$2" == "convert" ] || exit 1\n')
        fh.write('exec cp "$4" "$5"\n')
    os.chmod(fake, 0o755)

    with open(src / "registry.tdb", "w") as fh:
        fh.write("fake")
    with open(src / "passdb.tdb", "w") as fh:
        fh.write("fake")
    with open(src / "mango.tdb", "w") as fh:
        fh.write("fake")

    ctdb.migrate_tdb(None, str(dst))

    assert os.path.exists(dst / "registry.tdb.0")
    assert os.path.exists(dst / "passdb.tdb.0")
    assert not os.path.exists(dst / "mango.tdb.0")


def test_ensure_ctdbd_etc_files(tmpdir):
    src = tmpdir / "src"
    os.mkdir(src)
    dst = tmpdir / "dst"
    os.mkdir(dst)

    # this largely just creates a bunch of symlinks so it doesn't
    # need much fakery.
    ctdb.ensure_ctdbd_etc_files(etc_path=dst, src_path=src)
    assert os.path.islink(dst / "functions")
    assert os.path.islink(dst / "notify.sh")
    assert os.path.islink(dst / "events/legacy/00.ctdb.script")


def test_pnn_in_nodes(tmpdir):
    nodes_json = tmpdir / "nodes.json"
    real_path = tmpdir / "nodes"

    with pytest.raises(Exception):
        ctdb.pnn_in_nodes(0, nodes_json, real_path)

    with open(nodes_json, "w") as fh:
        fh.write("{}")
    result = ctdb.pnn_in_nodes(0, nodes_json, real_path)
    assert not result

    with open(nodes_json, "w") as fh:
        fh.write(
            """
            {"nodes": [
                {"node": "10.0.0.10", "pnn": 0, "state": "ready"},
                {"node": "10.0.0.11", "pnn": 1, "state": "new"}
            ]}
        """
        )
    result = ctdb.pnn_in_nodes(0, nodes_json, real_path)
    assert result
    result = ctdb.pnn_in_nodes(1, nodes_json, real_path)
    assert not result


class _Stop(Exception):
    pass


def test_manage_nodes(tmpdir, monkeypatch):
    nodes_json = tmpdir / "nodes.json"
    real_path = tmpdir / "nodes"
    monkeypatch.setattr(sambacc.samba_cmds, "_GLOBAL_PREFIX", ["true"])

    def once():
        raise _Stop()

    with pytest.raises(FileNotFoundError):
        ctdb.manage_nodes(
            0, nodes_json=nodes_json, real_path=real_path, pause_func=once
        )

    # node not present - can not update
    with open(nodes_json, "w") as fh:
        fh.write(
            """
            {"nodes": [
            ]}
        """
        )
    with pytest.raises(_Stop):
        ctdb.manage_nodes(
            0, nodes_json=nodes_json, real_path=real_path, pause_func=once
        )

    # node present, not in nodes - can not update
    with open(nodes_json, "w") as fh:
        fh.write(
            """
            {"nodes": [
                {"identity":"a", "node": "10.0.0.10", "pnn": 0, "state": "new"}
            ]}
        """
        )
    with pytest.raises(_Stop):
        ctdb.manage_nodes(
            0, nodes_json=nodes_json, real_path=real_path, pause_func=once
        )

    # node present, in nodes - nothing to do
    with open(nodes_json, "w") as fh:
        fh.write(
            """
            {"nodes": [
                {"identity":"a", "node": "10.0.0.10", "pnn": 0,
                 "state": "ready"}
            ]}
        """
        )
    with open(real_path, "w") as fh:
        fh.write("10.0.0.10\n")
    with pytest.raises(_Stop):
        ctdb.manage_nodes(
            0, nodes_json=nodes_json, real_path=real_path, pause_func=once
        )

    # node present, in nodes - new node in json
    with open(nodes_json, "w") as fh:
        fh.write(
            """
            {"nodes": [
                {"identity":"a", "node": "10.0.0.10", "pnn": 0,
                 "state": "ready"},
                {"identity":"b", "node": "10.0.0.11", "pnn": 1,
                 "state": "new"}
            ]}
        """
        )
    with open(real_path, "w") as fh:
        fh.write("10.0.0.10\n")
    with pytest.raises(_Stop):
        ctdb.manage_nodes(
            0, nodes_json=nodes_json, real_path=real_path, pause_func=once
        )
    with open(real_path, "r") as fh:
        lines = [x.strip() for x in fh.readlines()]
    assert "10.0.0.10" in lines
    assert "10.0.0.11" in lines

    # invalid state - nodes file and nodes json out of whack
    with open(nodes_json, "w") as fh:
        fh.write(
            """
            {"nodes": [
                {"identity":"a", "node": "10.0.0.10", "pnn": 0,
                 "state": "ready"},
                {"identity":"b", "node": "10.0.0.11", "pnn": 1,
                 "state": "new"}
            ]}
        """
        )
    with open(real_path, "w") as fh:
        fh.write("10.0.0.10\n")
        fh.write("10.0.0.12\n")
        fh.write("10.0.0.13\n")
    with pytest.raises(ValueError):
        ctdb.manage_nodes(
            0, nodes_json=nodes_json, real_path=real_path, pause_func=once
        )

    # node present but json file shows update incomplete
    with open(nodes_json, "w") as fh:
        fh.write(
            """
            {"nodes": [
                {"identity":"a", "node": "10.0.0.10", "pnn": 0,
                 "state": "ready"},
                {"identity":"b", "node": "10.0.0.11", "pnn": 1,
                 "state": "new"}
            ]}
        """
        )
    with open(real_path, "w") as fh:
        fh.write("10.0.0.10\n")
        fh.write("10.0.0.11\n")
    with pytest.raises(_Stop):
        ctdb.manage_nodes(
            0, nodes_json=nodes_json, real_path=real_path, pause_func=once
        )
    with open(real_path, "r") as fh:
        lines = [x.strip() for x in fh.readlines()]
    assert "10.0.0.10" in lines
    assert "10.0.0.11" in lines
    with open(nodes_json, "r") as fh:
        jdata = json.load(fh)
    assert jdata["nodes"][1]["node"] == "10.0.0.11"
    assert jdata["nodes"][1]["state"] == "ready"

    with open(nodes_json, "w") as fh:
        fh.write(
            """
            {"nodes": [
                {"identity":"a", "node": "10.0.0.10", "pnn": 0,
                 "state": "ready"},
                {"identity":"b", "node": "10.0.1.11", "pnn": 1,
                 "state": "changed"}
            ]}
        """
        )
    with open(real_path, "w") as fh:
        fh.write("10.0.0.10\n")
        fh.write("10.0.0.11\n")
    with pytest.raises(_Stop):
        ctdb.manage_nodes(
            0, nodes_json=nodes_json, real_path=real_path, pause_func=once
        )
    with open(real_path, "r") as fh:
        lines = [x.strip() for x in fh.readlines()]
    assert "10.0.0.10" in lines
    assert "#10.0.0.11" in lines
    with open(nodes_json, "r") as fh:
        jdata = json.load(fh)
    assert jdata["nodes"][1]["node"] == "10.0.1.11"
    assert jdata["nodes"][1]["state"] == "replaced"

    with pytest.raises(_Stop):
        ctdb.manage_nodes(
            0, nodes_json=nodes_json, real_path=real_path, pause_func=once
        )
    with open(real_path, "r") as fh:
        lines = [x.strip() for x in fh.readlines()]
    assert "10.0.0.10" in lines
    assert "10.0.1.11" in lines
    with open(nodes_json, "r") as fh:
        jdata = json.load(fh)
    assert jdata["nodes"][1]["node"] == "10.0.1.11"
    assert jdata["nodes"][1]["state"] == "ready"


def test_manage_nodes_refresh_fails(tmpdir, monkeypatch):
    nodes_json = tmpdir / "nodes.json"
    real_path = tmpdir / "nodes"
    monkeypatch.setattr(sambacc.samba_cmds, "_GLOBAL_PREFIX", ["false"])

    def once():
        raise _Stop()

    # node needs to be added
    with open(nodes_json, "w") as fh:
        fh.write(
            """
            {"nodes": [
                {"node": "10.0.0.10", "pnn": 0, "state": "ready"},
                {"node": "10.0.0.11", "pnn": 1, "state": "new"}
            ]}
        """
        )
    with open(real_path, "w") as fh:
        fh.write("10.0.0.10\n")
    with pytest.raises(Exception):
        ctdb.manage_nodes(
            0, nodes_json=nodes_json, real_path=real_path, pause_func=once
        )
    with open(real_path, "r") as fh:
        lines = [x.strip() for x in fh.readlines()]
    assert "10.0.0.10" in lines
    assert "10.0.0.11" in lines
    with open(nodes_json, "r") as fh:
        jdata = json.load(fh)
    assert jdata["nodes"][1]["node"] == "10.0.0.11"
    assert jdata["nodes"][1]["state"] == "new"


def test_manage_nodes_invalid_state(tmpdir):
    nodes_json = tmpdir / "nodes.json"
    real_path = tmpdir / "nodes"

    def once():
        raise _Stop()

    # node is ready but missing from nodes file
    with open(nodes_json, "w") as fh:
        fh.write(
            """
            {"nodes": [
                {"node": "10.0.0.10", "pnn": 0, "state": "ready"},
                {"node": "10.0.0.11", "pnn": 1, "state": "ready"}
            ]}
        """
        )
    with open(real_path, "w") as fh:
        fh.write("10.0.0.10\n")
    with pytest.raises(ValueError):
        ctdb.manage_nodes(
            0, nodes_json=nodes_json, real_path=real_path, pause_func=once
        )


def test_add_node_to_statefile(tmpdir):
    nodes_json = tmpdir / "nodes.json"

    ctdb.add_node_to_statefile(
        identity="node-0",
        node="10.0.0.10",
        pnn=0,
        path=nodes_json,
        in_nodes=True,
    )
    with open(nodes_json, "r") as fh:
        jdata = json.load(fh)
    assert jdata["nodes"][0]["node"] == "10.0.0.10"
    assert jdata["nodes"][0]["pnn"] == 0
    assert jdata["nodes"][0]["state"] == "ready"

    with pytest.raises(ValueError):
        ctdb.add_node_to_statefile(
            identity="node-0",
            node="10.0.0.11",
            pnn=0,
            path=nodes_json,
            in_nodes=False,
        )

    with pytest.raises(ValueError):
        ctdb.add_node_to_statefile(
            identity="node-0",
            node="10.0.1.11",
            pnn=2,
            path=nodes_json,
            in_nodes=False,
        )

    ctdb.add_node_to_statefile(
        identity="node-1",
        node="10.0.0.11",
        pnn=1,
        path=nodes_json,
        in_nodes=False,
    )
    with open(nodes_json, "r") as fh:
        jdata = json.load(fh)
    assert jdata["nodes"][0]["node"] == "10.0.0.10"
    assert jdata["nodes"][0]["pnn"] == 0
    assert jdata["nodes"][0]["state"] == "ready"
    assert jdata["nodes"][1]["node"] == "10.0.0.11"
    assert jdata["nodes"][1]["pnn"] == 1
    assert jdata["nodes"][1]["state"] == "new"


def test_ensure_ctdb_node_present(tmpdir):
    real_path = tmpdir / "nodes"
    lpath = tmpdir / "nodes.lnk"

    assert not os.path.exists(real_path)

    ctdb.ensure_ctdb_node_present(
        node="10.0.0.10", expected_pnn=0, real_path=real_path, canon_path=lpath
    )
    assert os.path.islink(lpath)
    with open(real_path, "r") as fh:
        lines = [x.strip() for x in fh.readlines()]
    assert ["10.0.0.10"] == lines

    ctdb.ensure_ctdb_node_present(
        node="10.0.0.11", expected_pnn=1, real_path=real_path, canon_path=lpath
    )
    assert os.path.islink(lpath)
    with open(real_path, "r") as fh:
        lines = [x.strip() for x in fh.readlines()]
    assert ["10.0.0.10", "10.0.0.11"] == lines

    with pytest.raises(ValueError):
        ctdb.ensure_ctdb_node_present(
            node="10.0.0.11",
            expected_pnn=0,
            real_path=real_path,
            canon_path=lpath,
        )


def test_write_ctdb_conf(tmpdir):
    path = tmpdir / "ctdb.conf"

    params = {
        "log_level": "DEBUG",
        "recovery_lock": "/tmp/foo/lock",
    }
    with open(path, "w") as fh:
        ctdb.write_ctdb_conf(fh, params)
    with open(path, "r") as fh:
        data = fh.read()
    assert "DEBUG" in data
    assert "/tmp/foo/lock" in data


def test_ensure_ctdb_conf(tmpdir):
    from .test_config import ctdb_config1
    from sambacc.config import GlobalConfig

    cfg = GlobalConfig(io.StringIO(ctdb_config1))
    path = tmpdir / "ctdb.conf"

    ctdb.ensure_ctdb_conf(iconfig=cfg.get("ctdb1"), path=path)
    with open(path, "r") as fh:
        data = fh.read()
    assert "DEBUG" in data
    assert "/var/lib/ctdb/shared/RECOVERY" in data


def test_ensure_smb_conf(tmpdir):
    from .test_config import ctdb_config1
    from sambacc.config import GlobalConfig

    cfg = GlobalConfig(io.StringIO(ctdb_config1))
    path = tmpdir / "smb.conf"

    ctdb.ensure_smb_conf(iconfig=cfg.get("ctdb1"), path=path)
    with open(path, "r") as fh:
        data = fh.read()
    assert "clustering = yes" in data
    assert "include = registry" in data


def test_refresh_node_in_statefile(tmpdir):
    nodes_json = tmpdir / "nodes.json"

    ctdb.add_node_to_statefile(
        identity="node-0",
        node="10.0.0.10",
        pnn=0,
        path=nodes_json,
        in_nodes=True,
    )

    # no changes
    ctdb.refresh_node_in_statefile(
        identity="node-0",
        node="10.0.0.10",
        pnn=0,
        path=nodes_json,
    )
    with open(nodes_json, "r") as fh:
        jdata = json.load(fh)
    assert len(jdata["nodes"]) == 1
    assert jdata["nodes"][0]["node"] == "10.0.0.10"
    assert jdata["nodes"][0]["identity"] == "node-0"
    assert jdata["nodes"][0]["pnn"] == 0
    assert jdata["nodes"][0]["state"] == "ready"

    # ip has changed
    ctdb.refresh_node_in_statefile(
        identity="node-0",
        node="10.0.1.10",
        pnn=0,
        path=nodes_json,
    )
    with open(nodes_json, "r") as fh:
        jdata = json.load(fh)
    assert len(jdata["nodes"]) == 1
    assert jdata["nodes"][0]["node"] == "10.0.1.10"
    assert jdata["nodes"][0]["identity"] == "node-0"
    assert jdata["nodes"][0]["pnn"] == 0
    assert jdata["nodes"][0]["state"] == "changed"

    with pytest.raises(ValueError):
        ctdb.refresh_node_in_statefile(
            identity="foobar",
            node="10.0.1.10",
            pnn=0,
            path=nodes_json,
        )

    with pytest.raises(ctdb.NodeNotPresent):
        ctdb.refresh_node_in_statefile(
            identity="node-1",
            node="10.0.0.11",
            pnn=1,
            path=nodes_json,
        )


def test_next_state():
    assert ctdb.next_state(ctdb.NodeState.READY) == ctdb.NodeState.READY
    assert ctdb.next_state(ctdb.NodeState.NEW) == ctdb.NodeState.READY
    assert ctdb.next_state(ctdb.NodeState.REPLACED) == ctdb.NodeState.READY
    assert ctdb.next_state(ctdb.NodeState.CHANGED) == ctdb.NodeState.REPLACED


def test_cli_leader_locator(tmpdir, monkeypatch, caplog):
    import logging

    caplog.set_level(logging.INFO)
    fake_ctdb = tmpdir / "fake_ctdb.sh"
    monkeypatch.setattr(sambacc.samba_cmds, "_GLOBAL_PREFIX", [fake_ctdb])
    monkeypatch.setenv("SAMBA_SPECIFICS", "ctdb_leader_admin_command")
    ldr_admin_cmd = sambacc.samba_cmds.ctdb_leader_admin_cmd()

    def _fake_ctdb_script(pnn, recmaster):
        with open(fake_ctdb, "w") as fh:
            fh.write("#!/bin/sh\n")
            fh.write("case $2 in\n")
            fh.write(f"pnn) {pnn};;\n")
            fh.write(f"{ldr_admin_cmd}) {recmaster};;\n")
            fh.write("esac\n")
            fh.write("exit 5\n")
        os.chmod(fake_ctdb, 0o700)

    _fake_ctdb_script(pnn="echo 0; exit 0", recmaster="echo 0; exit 0")
    with ctdb.CLILeaderLocator() as status:
        assert status.is_leader()

    _fake_ctdb_script(pnn="echo 1; exit 0", recmaster="echo 0; exit 0")
    with ctdb.CLILeaderLocator() as status:
        assert not status.is_leader()

    # test error handling
    _fake_ctdb_script(pnn="exit 1", recmaster="echo 0; exit 0")
    with ctdb.CLILeaderLocator() as status:
        assert not status.is_leader()
    assert "pnn" in caplog.records[-1].getMessage()
    assert "['" + ldr_admin_cmd + "']" not in caplog.records[-1].getMessage()
    _fake_ctdb_script(pnn="echo 1; exit 0", recmaster="exit 1")
    with ctdb.CLILeaderLocator() as status:
        assert not status.is_leader()
    assert "pnn" not in caplog.records[-1].getMessage()
    assert "['" + ldr_admin_cmd + "']" in caplog.records[-1].getMessage()

    os.unlink(fake_ctdb)
    with ctdb.CLILeaderLocator() as status:
        assert not status.is_leader()
    assert "pnn" in caplog.records[-2].getMessage()
    assert "['" + ldr_admin_cmd + "']" in caplog.records[-1].getMessage()


def test_check_nodestatus(tmp_path):
    import os

    datapath = tmp_path / "_ctdb"
    datapath.mkdir()

    fake_ctdb = [
        "#!/bin/sh",
        'if [ "$1$TESTFAIL" == "nodestatus" ]',
        "then exit 0;",
        "else exit 1;",
        "fi",
    ]
    fake_ctdb_script = datapath / "ctdb.sh"
    with open(fake_ctdb_script, "w") as fh:
        fh.write("\n".join(fake_ctdb))
        fh.write("\n")
    os.chmod(fake_ctdb_script, 0o755)

    test_cmd = sambacc.samba_cmds.SambaCommand(fake_ctdb_script)
    # simulate nodestatus == OK
    pid = os.fork()
    if pid == 0:
        ctdb.check_nodestatus(cmd=test_cmd)
    else:
        _, status = os.waitpid(pid, 0)
        assert status == 0

    # simulate nodestatus != OK
    pid = os.fork()
    if pid == 0:
        os.environ["TESTFAIL"] = "yes"
        ctdb.check_nodestatus(cmd=test_cmd)
    else:
        _, status = os.waitpid(pid, 0)
        assert status != 0



#
# sambacc: a samba container configuration tool
# Copyright (C) 2021  John Mulligan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>
#

import io
import pytest

import sambacc.config
import sambacc.netcmd_loader
import sambacc.samba_cmds

smb_conf = """
[global]
cache directory = {path}
state directory = {path}
private dir = {path}
include = registry
"""

config1 = """
{
  "samba-container-config": "v0",
  "configs": {
    "foobar": {
      "shares": [
        "share",
        "stuff"
      ],
      "globals": ["global0"],
      "instance_name": "GANDOLPH"
    }
  },
  "shares": {
    "share": {
      "options": {
        "path": "/share",
        "read only": "no",
        "valid users": "sambauser",
        "guest ok": "no",
        "force user": "root"
      }
    },
    "stuff": {
      "options": {
        "path": "/mnt/stuff"
      }
    }
  },
  "globals": {
    "global0": {
      "options": {
        "workgroup": "SAMBA",
        "security": "user",
        "server min protocol": "SMB2",
        "load printers": "no",
        "printing": "bsd",
        "printcap name": "/dev/null",
        "disable spoolss": "yes",
        "guest ok": "no"
      }
    }
  },
  "_extra_junk": 0
}
"""


@pytest.fixture(scope="function")
def testloader(tmp_path):
    data_path = tmp_path / "_samba"
    data_path.mkdir()
    smb_conf_path = tmp_path / "smb.conf"
    with open(smb_conf_path, "w") as fh:
        fh.write(smb_conf.format(path=data_path))

    ldr = sambacc.netcmd_loader.NetCmdLoader()
    ldr._net_conf = sambacc.samba_cmds.net[
        "--configfile={}".format(smb_conf_path), "conf"
    ]
    return ldr


def test_import(testloader):
    fh = io.StringIO(config1)
    g = sambacc.config.GlobalConfig(fh)
    testloader.import_config(g.get("foobar"))


def test_current_shares(testloader):
    shares = testloader.current_shares()
    assert len(shares) == 0
    fh = io.StringIO(config1)
    g = sambacc.config.GlobalConfig(fh)
    testloader.import_config(g.get("foobar"))
    shares = testloader.current_shares()
    assert len(shares) == 2
    assert "share" in shares
    assert "stuff" in shares


def test_dump(testloader, tmp_path):
    fh = io.StringIO(config1)
    g = sambacc.config.GlobalConfig(fh)
    testloader.import_config(g.get("foobar"))

    with open(tmp_path / "dump.txt", "w") as fh:
        testloader.dump(fh)
    with open(tmp_path / "dump.txt") as fh:
        dump = fh.read()

    assert "[global]" in dump
    assert "netbios name = GANDOLPH" in dump
    assert "[share]" in dump
    assert "path = /share" in dump
    assert "[stuff]" in dump
    assert "path = /mnt/stuff" in dump


def test_set(testloader, tmp_path):
    testloader.set("global", "client signing", "mandatory")

    with open(tmp_path / "dump.txt", "w") as fh:
        testloader.dump(fh)
    with open(tmp_path / "dump.txt") as fh:
        dump = fh.read()

    assert "[global]" in dump
    assert "client signing = mandatory" in dump


def test_loader_error_set(testloader, tmp_path):
    with pytest.raises(sambacc.netcmd_loader.LoaderError):
        testloader.set("", "", "yikes")



#
# sambacc: a samba container configuration tool
# Copyright (C) 2022  John Mulligan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>
#
"""xattr shim module

This module exists to insulate sambacc from the platform xattr module.
Currently it only support pyxattr. This module can be imported without
pyxattr (xattr) present. The functions will import the required module
and raise an ImportError if xattr is not available.

This shim also provides a typed functions for xattr management. This
could have been accomplished by writing a pyi file for xattr but since
we need the runtime support we just add new functions.
"""


import pathlib
import typing

XAttrItem = typing.Union[
    int,  # an open file descriptor, not wrapped by an object
    pathlib.Path,  # pathlib path object
    str,  # basic path string
    typing.IO,  # an open file descriptor, wrapped by an object
]
Namespace = typing.Optional[bytes]


def get(
    item: XAttrItem,
    name: str,
    *,
    nofollow: bool = False,
    namespace: Namespace = None
) -> bytes:
    """Get an xattr from the target item and name.
    See docs for PyXattr module for details.
    """
    import xattr  # type: ignore

    kwargs: dict[str, typing.Any] = {"nofollow": nofollow}
    if namespace is not None:
        kwargs["namespace"] = namespace
    return xattr.get(item, name, **kwargs)


def set(
    item: XAttrItem,
    name: str,
    value: str,
    *,
    flags: typing.Optional[int] = None,
    nofollow: bool = False,
    namespace: Namespace = None
) -> None:
    """Set an xattr. See docs for PyXattr module for details."""
    import xattr  # type: ignore

    kwargs: dict[str, typing.Any] = {"nofollow": nofollow}
    if flags is not None:
        kwargs["flags"] = flags
    if namespace is not None:
        kwargs["namespace"] = namespace
    return xattr.set(item, name, value, **kwargs)



#
# sambacc: a samba container configuration tool
# Copyright (C) 2021  John Mulligan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>
#

from sambacc import samba_cmds
import sambacc.paths as paths

from .cli import commands, Context, Fail
from .initialize import init_container, setup_step_names
from .join import join


INIT_ALL = "init-all"


def _run_container_args(parser):
    parser.add_argument(
        "--no-init",
        action="store_true",
        help=(
            "(DEPRECATED - see --setup) Do not initialize the container"
            " envionment. Only start running the target process."
        ),
    )
    _setup_choices = [INIT_ALL] + list(setup_step_names())
    parser.add_argument(
        "--setup",
        action="append",
        choices=_setup_choices,
        help=(
            "Specify one or more setup step names to preconfigure the"
            " container environment before the server process is started."
            " The special 'init-all' name will perform all known setup steps."
        ),
    )
    parser.add_argument(
        "--insecure-auto-join",
        action="store_true",
        help=(
            "Perform an inscure domain join prior to starting a service."
            " Based on env vars JOIN_USERNAME and INSECURE_JOIN_PASSWORD."
        ),
    )
    parser.add_argument(
        "target",
        choices=["smbd", "winbindd", "ctdbd"],
        help="Which process to run",
    )


@commands.command(name="run", arg_func=_run_container_args)
def run_container(ctx: Context) -> None:
    """Run a specified server process."""
    if ctx.cli.no_init and ctx.cli.setup:
        raise Fail("can not specify both --no-init and --setup")
    # running servers expect to make use of ctdb whenever it is configured
    ctx.expects_ctdb = True
    if not ctx.cli.no_init and not ctx.cli.setup:
        # TODO: drop this along with --no-init and move to a opt-in
        # rather than opt-out form of pre-run setup
        init_container(ctx)
    elif ctx.cli.setup:
        steps = list(ctx.cli.setup)
        init_container(ctx, steps=(None if INIT_ALL in steps else steps))

    paths.ensure_samba_dirs()
    if ctx.cli.target == "smbd":
        # execute smbd process
        samba_cmds.execute(samba_cmds.smbd_foreground())
    elif ctx.cli.target == "winbindd":
        if getattr(ctx.cli, "insecure_auto_join", False):
            join(ctx)
        # execute winbind process
        samba_cmds.execute(samba_cmds.winbindd_foreground())
    elif ctx.cli.target == "ctdbd":
        samba_cmds.execute(samba_cmds.ctdbd_foreground)
    else:
        raise Fail(f"invalid target process: {ctx.cli.target}")



#
# sambacc: a samba container configuration tool
# Copyright (C) 2021  John Mulligan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>
#

import sambacc.simple_waiter


def test_generate_sleeps():
    g = sambacc.simple_waiter.generate_sleeps()
    times = [next(g) for _ in range(130)]
    assert times[0] == 1
    assert times[0:11] == [1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1]
    assert times[12] == 5
    assert all(times[x] == 5 for x in range(12, 33))
    assert times[34] == 60
    assert all(times[x] == 60 for x in range(34, 130))


def test_sleeper():
    def gen():
        while True:
            yield 8

    cc = 0

    def fake_sleep(v):
        nonlocal cc
        cc += 1
        assert v == 8

    sleeper = sambacc.simple_waiter.Sleeper(times=gen())
    sleeper._sleep = fake_sleep
    sleeper.wait()
    assert cc == 1
    for _ in range(3):
        sleeper.wait()
    assert cc == 4

    cc = 0

    def fake_sleep2(v):
        nonlocal cc
        cc += 1
        assert v == 1

    sleeper = sambacc.simple_waiter.Sleeper()
    sleeper._sleep = fake_sleep2
    sleeper.wait()
    assert cc == 1
    for _ in range(3):
        sleeper.wait()
    assert cc == 4



#
# sambacc: a samba container configuration tool
# Copyright (C) 2022  John Mulligan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>
#

from __future__ import annotations

import contextlib
import datetime
import errno
import logging
import os
import typing

from sambacc import _xattr as xattr


_logger = logging.getLogger(__name__)


class PermissionsHandler(typing.Protocol):
    def has_status(self) -> bool:
        """Return true if the path has status metadata."""
        ...  # pragma: no cover

    def status_ok(self) -> bool:
        """Return true if status is OK (no changes are needed)."""
        ...  # pragma: no cover

    def update(self) -> None:
        """Update the permissions as needed."""
        ...  # pragma: no cover

    def path(self) -> str:
        """Return the path under consideration."""
        ...  # pragma: no cover


@contextlib.contextmanager
def _opendir(path: str) -> typing.Iterator[int]:
    dfd: int = os.open(path, os.O_DIRECTORY)
    try:
        yield dfd
        os.fsync(dfd)
    except OSError:
        os.sync()
        raise
    finally:
        os.close(dfd)


class NoopPermsHandler:
    def __init__(
        self,
        path: str,
        status_xattr: str,
        options: typing.Dict[str, str],
        root: str = "/",
    ) -> None:
        self._path = path

    def path(self) -> str:
        return self._path

    def has_status(self) -> bool:
        return False

    def status_ok(self) -> bool:
        return True

    def update(self) -> None:
        pass


class InitPosixPermsHandler:
    """Initialize posix permissions on a share (directory).

    This handler sets posix permissions only.

    It will only set the permissions when the status xattr does not
    match the expected prefix value. This prevents it from overwiting
    permissions that may have been changed intentionally after
    share initialization.
    """

    _default_mode = 0o777
    _default_status_prefix = "v1"

    def __init__(
        self,
        path: str,
        status_xattr: str,
        options: typing.Dict[str, str],
        root: str = "/",
    ) -> None:
        self._path = path
        self._root = root
        self._xattr = status_xattr
        try:
            self._mode = int(options["mode"], 8)
        except KeyError:
            self._mode = self._default_mode
        try:
            self._prefix = options["status_prefix"]
        except KeyError:
            self._prefix = self._default_status_prefix

    def path(self) -> str:
        return self._path

    def _full_path(self) -> str:
        return os.path.join(self._root, self._path.lstrip("/"))

    def has_status(self) -> bool:
        try:
            self._get_status()
            return True
        except KeyError:
            return False

    def status_ok(self) -> bool:
        try:
            sval = self._get_status()
        except KeyError:
            return False
        curr_prefix = sval.split("/")[0]
        return curr_prefix == self._prefix

    def update(self) -> None:
        if self.status_ok():
            return
        self._set_perms()
        self._set_status()

    def _get_status(self) -> str:
        path = self._full_path()
        _logger.debug("reading xattr %r: %r", self._xattr, path)
        try:
            value = xattr.get(path, self._xattr, nofollow=True)
        except OSError as err:
            if err.errno == errno.ENODATA:
                raise KeyError(self._xattr)
            raise
        return value.decode("utf8")

    def _set_perms(self) -> None:
        # yeah, this is really simple compared to all the state management
        # stuff.
        path = self._full_path()
        with _opendir(path) as dfd:
            os.fchmod(dfd, self._mode)

    def _timestamp(self) -> str:
        return datetime.datetime.now().strftime("%s")

    def _set_status(self) -> None:
        # we save the marker prefix followed by a timestamp as a debugging hint
        ts = self._timestamp()
        val = f"{self._prefix}/{ts}"
        path = self._full_path()
        _logger.debug("setting xattr %r=%r: %r", self._xattr, val, self._path)
        with _opendir(path) as dfd:
            xattr.set(dfd, self._xattr, val, nofollow=True)


class AlwaysPosixPermsHandler(InitPosixPermsHandler):
    """Works like the init handler, but always sets the permissions,
    even if the status xattr exists and is valid.
    May be useful for testing and debugging.
    """

    def update(self) -> None:
        self._set_perms()
        self._set_status()



#
# sambacc: a samba container configuration tool
# Copyright (C) 2022  John Mulligan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>
#

import os

import pytest


import sambacc.permissions


@pytest.mark.parametrize(
    "cls",
    [
        sambacc.permissions.NoopPermsHandler,
        sambacc.permissions.InitPosixPermsHandler,
        sambacc.permissions.AlwaysPosixPermsHandler,
    ],
)
def test_permissions_path(cls):
    assert cls("/foo", "user.foo", options={}).path() == "/foo"


def test_noop_handler():
    nh = sambacc.permissions.NoopPermsHandler("/foo", "user.foo", options={})
    assert nh.path() == "/foo"
    assert not nh.has_status()
    assert nh.status_ok()
    assert nh.update() is None


@pytest.fixture(scope="function")
def tmp_path_xattrs_ok(tmp_path_factory):
    try:
        import xattr  # type: ignore
    except ImportError:
        pytest.skip("xattr module not available")

    tmpp = tmp_path_factory.mktemp("needs_xattrs")
    try:
        xattr.set(str(tmpp), "user.deleteme", "1")
        xattr.remove(str(tmpp), "user.deleteme")
    except OSError:
        raise pytest.skip(
            "temp dir does not support xattrs"
            " (try changing basetmp to a file system that supports xattrs)"
        )
    return tmpp


def test_init_handler(tmp_path_xattrs_ok):
    path = tmp_path_xattrs_ok / "foo"
    os.mkdir(path)
    ih = sambacc.permissions.InitPosixPermsHandler(
        str(path), "user.marker", options={}
    )
    assert ih.path().endswith("/foo")
    assert not ih.has_status()
    assert not ih.status_ok()

    ih.update()
    assert ih.has_status()
    assert ih.status_ok()

    os.chmod(path, 0o755)
    ih.update()
    assert (os.stat(path).st_mode & 0o777) == 0o755


def test_always_handler(tmp_path_xattrs_ok):
    path = tmp_path_xattrs_ok / "foo"
    os.mkdir(path)
    ih = sambacc.permissions.AlwaysPosixPermsHandler(
        str(path), "user.marker", options={}
    )
    assert ih.path().endswith("/foo")
    assert not ih.has_status()
    assert not ih.status_ok()

    ih.update()
    assert ih.has_status()
    assert ih.status_ok()

    os.chmod(path, 0o755)
    assert (os.stat(path).st_mode & 0o777) == 0o755
    ih.update()
    assert (os.stat(path).st_mode & 0o777) == 0o777



#
# sambacc: a samba container configuration tool
# Copyright (C) 2021  John Mulligan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>
#

import sambacc.samba_cmds


def test_create_samba_command():
    cmd = sambacc.samba_cmds.SambaCommand("hello")
    assert cmd.name == "hello"
    cmd2 = cmd["world"]
    assert cmd.name == "hello"
    assert list(cmd) == ["hello"]
    assert list(cmd2) == ["hello", "world"]


def test_debug_command():
    cmd = sambacc.samba_cmds.SambaCommand("beep", debug="5")
    assert list(cmd) == ["beep", "--debuglevel=5"]


def test_global_debug():
    sambacc.samba_cmds.set_global_debug("7")
    try:
        cmd = sambacc.samba_cmds.SambaCommand("cheep")
        assert list(cmd) == ["cheep", "--debuglevel=7"]
    finally:
        sambacc.samba_cmds.set_global_debug("")


def test_global_prefix():
    # enabled
    sambacc.samba_cmds.set_global_prefix(["bob"])
    try:
        cmd = sambacc.samba_cmds.SambaCommand("deep")
        assert list(cmd) == ["bob", "deep"]
        assert cmd.name == "bob"
    finally:
        sambacc.samba_cmds.set_global_prefix([])

    # disabled
    cmd = sambacc.samba_cmds.SambaCommand("deep")
    assert list(cmd) == ["deep"]
    assert cmd.name == "deep"


def test_global_prefix_extended():
    # enabled
    sambacc.samba_cmds.set_global_prefix(["frank"])
    try:
        cmd = sambacc.samba_cmds.SambaCommand("deep")[
            "13", "--future=not-too-distant"
        ]
        assert list(cmd) == ["frank", "deep", "13", "--future=not-too-distant"]
        assert cmd.name == "frank"
    finally:
        sambacc.samba_cmds.set_global_prefix([])

    # disabled, must not "inherit" the prefix
    cmd2 = cmd["--scheme", "evil"]
    assert list(cmd2) == [
        "deep",
        "13",
        "--future=not-too-distant",
        "--scheme",
        "evil",
    ]
    assert cmd2.name == "deep"


def test_command_repr():
    cmd = sambacc.samba_cmds.SambaCommand("doop")
    cr = repr(cmd)
    assert cr.startswith("SambaCommand")
    assert "doop" in cr


def test_encode_none():
    res = sambacc.samba_cmds.encode(None)
    assert res == b""


def test_execute():
    import os

    cmd = sambacc.samba_cmds.SambaCommand("true")
    pid = os.fork()
    if pid == 0:
        sambacc.samba_cmds.execute(cmd)
    else:
        _, status = os.waitpid(pid, 0)
        assert status == 0


def test_create_command_args():
    # this is the simpler base class for SambaCommand. It lacks
    # the samba debug level option.
    cmd = sambacc.samba_cmds.CommandArgs("something")
    assert cmd.name == "something"
    cmd2 = cmd["nice"]
    assert cmd.name == "something"
    assert list(cmd) == ["something"]
    assert list(cmd2) == ["something", "nice"]


def test_command_args_repr():
    r = str(sambacc.samba_cmds.CommandArgs("something", ["nice"]))
    assert r.startswith("CommandArgs")
    assert "something" in r
    assert "nice" in r


def test_get_samba_specifics(monkeypatch):
    monkeypatch.setenv("SAMBA_SPECIFICS", "")
    ss = sambacc.samba_cmds.get_samba_specifics()
    assert not ss

    monkeypatch.setenv("SAMBA_SPECIFICS", "wibble,quux")
    ss = sambacc.samba_cmds.get_samba_specifics()
    assert ss
    assert len(ss) == 2
    assert "wibble" in ss
    assert "quux" in ss


def test_smbd_foreground(monkeypatch):
    monkeypatch.setenv("SAMBA_SPECIFICS", "")
    sf = sambacc.samba_cmds.smbd_foreground()
    assert "smbd" in sf.name
    assert "--log-stdout" in sf.argv()
    assert "--debug-stdout" not in sf.argv()

    monkeypatch.setenv("SAMBA_SPECIFICS", "daemon_cli_debug_output")
    sf = sambacc.samba_cmds.smbd_foreground()
    assert "smbd" in sf.name
    assert "--log-stdout" not in sf.argv()
    assert "--debug-stdout" in sf.argv()


def test_winbindd_foreground(monkeypatch):
    monkeypatch.setenv("SAMBA_SPECIFICS", "")
    wf = sambacc.samba_cmds.winbindd_foreground()
    assert "winbindd" in wf.name
    assert "--stdout" in wf.argv()
    assert "--debug-stdout" not in wf.argv()

    monkeypatch.setenv("SAMBA_SPECIFICS", "daemon_cli_debug_output")
    wf = sambacc.samba_cmds.winbindd_foreground()
    assert "winbindd" in wf.name
    assert "--stdout" not in wf.argv()
    assert "--debug-stdout" in wf.argv()



import io

import sambacc.passwd_loader
from .test_config import config2

etc_passwd1 = """
root:x:0:0:root:/root:/bin/bash
bin:x:1:1:bin:/bin:/sbin/nologin
daemon:x:2:2:daemon:/sbin:/sbin/nologin
adm:x:3:4:adm:/var/adm:/sbin/nologin
lp:x:4:7:lp:/var/spool/lpd:/sbin/nologin
sync:x:5:0:sync:/sbin:/bin/sync
shutdown:x:6:0:shutdown:/sbin:/sbin/shutdown
halt:x:7:0:halt:/sbin:/sbin/halt
mail:x:8:12:mail:/var/spool/mail:/sbin/nologin
operator:x:11:0:operator:/root:/sbin/nologin
games:x:12:100:games:/usr/games:/sbin/nologin
ftp:x:14:50:FTP User:/var/ftp:/sbin/nologin
nobody:x:65534:65534:Kernel Overflow User:/:/sbin/nologin
dbus:x:81:81:System message bus:/:/sbin/nologin
""".strip()

etc_group1 = """
root:x:0:
bin:x:1:
daemon:x:2:
sys:x:3:
adm:x:4:
tty:x:5:
disk:x:6:
lp:x:7:
mem:x:8:
kmem:x:9:
wheel:x:10:
cdrom:x:11:
mail:x:12:
man:x:15:
dialout:x:18:
floppy:x:19:
games:x:20:
tape:x:33:
video:x:39:
ftp:x:50:
lock:x:54:
audio:x:63:
users:x:100:
nobody:x:65534:
utmp:x:22:
utempter:x:35:
kvm:x:36:
dbus:x:81:
""".strip()


def test_read_existing_passwd():
    fh = io.StringIO(etc_passwd1)
    pfl = sambacc.passwd_loader.PasswdFileLoader()
    pfl.readfp(fh)
    assert len(pfl.lines) == 14
    assert pfl.lines[0].startswith("root")
    fh2 = io.StringIO()
    pfl.writefp(fh2)
    assert etc_passwd1 == fh2.getvalue()


def test_read_existing_group():
    fh = io.StringIO(etc_group1)
    pfl = sambacc.passwd_loader.GroupFileLoader()
    pfl.readfp(fh)
    assert len(pfl.lines) == 28
    assert pfl.lines[0].startswith("root")
    fh2 = io.StringIO()
    pfl.writefp(fh2)
    assert etc_group1 == fh2.getvalue()


def test_add_user():
    fh = io.StringIO(config2)
    g = sambacc.config.GlobalConfig(fh)
    ic = g.get("foobar")
    users = list(ic.users())

    pfl = sambacc.passwd_loader.PasswdFileLoader()
    for u in users:
        pfl.add_user(u)
    assert len(pfl.lines) == 3
    fh2 = io.StringIO()
    pfl.writefp(fh2)
    txt = fh2.getvalue()
    assert "alice:x:" in txt
    assert "bob:x:" in txt
    assert "carol:x:" in txt


def test_add_group():
    fh = io.StringIO(config2)
    g = sambacc.config.GlobalConfig(fh)
    ic = g.get("foobar")
    groups = list(ic.groups())

    gfl = sambacc.passwd_loader.GroupFileLoader()
    for g in groups:
        gfl.add_group(g)
    # test that duplicates don't add extra lines
    for g in groups:
        gfl.add_group(g)
    assert len(gfl.lines) == 3
    fh2 = io.StringIO()
    gfl.writefp(fh2)
    txt = fh2.getvalue()
    assert "alice:x:" in txt
    assert "bob:x:" in txt
    assert "carol:x:" in txt


def test_read_passwd_file(tmp_path):
    fname = tmp_path / "read_etc_passwd"
    with open(fname, "w") as fh:
        fh.write(etc_passwd1)
    pfl = sambacc.passwd_loader.PasswdFileLoader(fname)
    pfl.read()
    assert len(pfl.lines) == 14
    assert pfl.lines[0].startswith("root")
    fh2 = io.StringIO()
    pfl.writefp(fh2)
    assert etc_passwd1 == fh2.getvalue()


def test_write_passwd_file(tmp_path):
    fh = io.StringIO(config2)
    g = sambacc.config.GlobalConfig(fh)
    ic = g.get("foobar")
    users = list(ic.users())

    fname = tmp_path / "write_etc_passwd"
    with open(fname, "w") as fh:
        fh.write(etc_passwd1)

    pfl = sambacc.passwd_loader.PasswdFileLoader(fname)
    pfl.read()
    for u in users:
        pfl.add_user(u)
    # test that duplicates don't add extra lines
    for u in users:
        pfl.add_user(u)
    assert len(pfl.lines) == 17
    pfl.write()

    with open(fname) as fh:
        txt = fh.read()
    assert "root:x:" in txt
    assert "\nalice:x:" in txt
    assert "\nbob:x:" in txt
    assert "\ncarol:x:" in txt



#
# sambacc: a samba container configuration tool
# Copyright (C) 2021  John Mulligan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>
#

from __future__ import annotations

import os
import typing

DebugLevel = typing.Optional[str]
ArgList = typing.Optional[list[str]]

_GLOBAL_PREFIX: list[str] = []
_GLOBAL_DEBUG: str = ""


# Known flags for SAMBA_SPECIFICS env variable
_DAEMON_CLI_STDOUT_OPT: str = "daemon_cli_debug_output"
_CTDB_LEADER_ADMIN_CMD: str = "ctdb_leader_admin_command"


def get_samba_specifics() -> typing.Set[str]:
    value = os.environ.get("SAMBA_SPECIFICS", "")
    out = set()
    if value:
        for v in value.split(","):
            out.add(v)
    return out


def _daemon_stdout_opt(daemon: str) -> str:
    if daemon == "smbd":
        opt = "--log-stdout"
    else:
        opt = "--stdout"
    opt_lst = get_samba_specifics()
    if _DAEMON_CLI_STDOUT_OPT in opt_lst:
        opt = "--debug-stdout"
    return opt


def ctdb_leader_admin_cmd() -> str:
    leader_cmd = "recmaster"
    opt_lst = get_samba_specifics()
    if _CTDB_LEADER_ADMIN_CMD in opt_lst:
        leader_cmd = "leader"
    return leader_cmd


def set_global_prefix(lst: list[str]) -> None:
    _GLOBAL_PREFIX[:] = lst


def set_global_debug(level: str) -> None:
    global _GLOBAL_DEBUG
    _GLOBAL_DEBUG = level


def _to_args(value: typing.Any) -> list[str]:
    if isinstance(value, str):
        return [value]
    return [str(v) for v in value]


class CommandArgs:
    """A utility class for building command line commands."""

    _name: str
    args: list[str]
    cmd_prefix: list[str]

    def __init__(self, name: str, args: ArgList = None):
        self._name = name
        self.args = args or []
        self.cmd_prefix = []

    def __getitem__(self, new_value: typing.Any) -> CommandArgs:
        return self.__class__(self._name, args=self.args + _to_args(new_value))

    def raw_args(self) -> list[str]:
        return [self._name] + self.args

    def prefix_args(self) -> list[str]:
        return list(_GLOBAL_PREFIX) + list(self.cmd_prefix)

    def argv(self) -> list[str]:
        return self.prefix_args() + self.raw_args()

    def __iter__(self) -> typing.Iterator[str]:
        return iter(self.argv())

    def __repr__(self) -> str:
        return "CommandArgs({!r}, {!r})".format(self._name, self.args)

    @property
    def name(self) -> str:
        """Return the command to be executed. This may differ from
        the underlying command.
        """
        return self.argv()[0]


class SambaCommand(CommandArgs):
    """A utility class for building samba (or any) command line command."""

    debug: DebugLevel

    def __init__(
        self, name: str, args: ArgList = None, debug: DebugLevel = None
    ):
        super().__init__(name, args)
        self.debug = debug

    def __getitem__(self, new_value: typing.Any) -> SambaCommand:
        return self.__class__(
            self._name,
            args=self.args + _to_args(new_value),
            debug=self.debug,
        )

    def _debug_args(self, dlvl: str = "--debuglevel={}") -> list[str]:
        if self.debug:
            return [dlvl.format(self.debug)]
        if _GLOBAL_DEBUG:
            return [dlvl.format(_GLOBAL_DEBUG)]
        return []

    def raw_args(self) -> list[str]:
        return [self._name] + self._debug_args() + self.args

    def __repr__(self) -> str:
        return "SambaCommand({!r}, {!r}, {!r})".format(
            self._name, self.args, self.debug
        )


net = SambaCommand("net")

wbinfo = SambaCommand("wbinfo")

smbd = SambaCommand("/usr/sbin/smbd")

winbindd = SambaCommand("/usr/sbin/winbindd")

samba_dc = SambaCommand("/usr/sbin/samba")


def smbd_foreground() -> SambaCommand:
    return smbd[
        "--foreground", _daemon_stdout_opt("smbd"), "--no-process-group"
    ]


def winbindd_foreground() -> SambaCommand:
    return winbindd[
        "--foreground", _daemon_stdout_opt("winbindd"), "--no-process-group"
    ]


def samba_dc_foreground() -> SambaCommand:
    return samba_dc["--foreground"]


ctdbd = SambaCommand("/usr/sbin/ctdbd")

ctdbd_foreground = ctdbd["--interactive"]

ltdbtool = CommandArgs("ltdbtool")

ctdb = SambaCommand("ctdb")

sambatool = SambaCommand("samba-tool")

smbcontrol = SambaCommand("smbcontrol")


def encode(value: typing.Union[str, bytes, None]) -> bytes:
    if value is None:
        return b""
    elif isinstance(value, str):
        value = value.encode("utf8")
    return value


def execute(cmd: SambaCommand) -> None:
    """Exec into the command specified (without forking)."""
    os.execvp(cmd.name, cmd.argv())  # pragma: no cover



import contextlib
import io
import os
import pytest
import shutil

import sambacc.passdb_loader
import sambacc.config
from .test_config import config2

_smb_conf = """
[global]
cache directory = {path}
state directory = {path}
private dir = {path}
include = registry
"""

passwd_append1 = [
    "alice:x:1010:1010:ALICE:/home/alice:/bin/bash\n",
    "bob:x:1011:1011:BOB:/home/bob:/bin/bash\n",
    "carol:x:1010:1010:carol:/home/alice:/bin/bash\n",
]


@pytest.fixture(scope="function")
def smb_conf(tmp_path):
    data_path = tmp_path / "_samba"
    data_path.mkdir()
    smb_conf_path = tmp_path / "smb.conf"
    with open(smb_conf_path, "w") as fh:
        fh.write(_smb_conf.format(path=data_path))
    return smb_conf_path


@contextlib.contextmanager
def alter_passwd(path, append):
    bkup = path / "passwd.bak"
    mypasswd = os.environ.get("NSS_WRAPPER_PASSWD")
    shutil.copy(mypasswd, bkup)
    with open(mypasswd, "a") as fh:
        fh.write("\n")
        for line in passwd_append1:
            fh.write(line)
    yield
    shutil.copy(bkup, mypasswd)


def requires_passdb_modules():
    try:
        sambacc.passdb_loader._samba_modules()
    except ImportError:
        pytest.skip("unable to load samba passdb modules")


def test_init_custom_smb_conf(smb_conf):
    requires_passdb_modules()
    sambacc.passdb_loader.PassDBLoader(smbconf=str(smb_conf))


def test_init_default_smb_conf():
    requires_passdb_modules()
    # this is a bit hacky, but I don't want to assume the local
    # system has or doesn't have a "real" smb.conf
    if os.path.exists("/etc/samba/smb.conf"):
        sambacc.passdb_loader.PassDBLoader(smbconf=None)
    else:
        with pytest.raises(Exception):
            sambacc.passdb_loader.PassDBLoader(smbconf=None)


def test_add_users(tmp_path, smb_conf):
    requires_passdb_modules()
    # TODO: actually use nss_wrapper!
    if not os.environ.get("NSS_WRAPPER_PASSWD"):
        pytest.skip("need to have path to passwd file")
    if os.environ.get("WRITABLE_PASSWD") != "yes":
        pytest.skip("need to append users to passwd file")
    with alter_passwd(tmp_path, passwd_append1):
        fh = io.StringIO(config2)
        g = sambacc.config.GlobalConfig(fh)
        ic = g.get("foobar")
        users = list(ic.users())

        pdbl = sambacc.passdb_loader.PassDBLoader(smbconf=str(smb_conf))
        for u in users:
            pdbl.add_user(u)


def test_add_user_not_in_passwd(smb_conf):
    requires_passdb_modules()
    pdbl = sambacc.passdb_loader.PassDBLoader(smbconf=str(smb_conf))

    # Irritatingly, the passwd file contents appear to be cached
    # so we need to make up a user that is def. not in the etc passwd
    # equivalent, in order to get samba libs to reject it
    urec = dict(name="nogoodnik", uid=101010, gid=101010, password="yuck")
    ubad = sambacc.config.UserEntry(None, urec, 0)
    with pytest.raises(Exception):
        pdbl.add_user(ubad)


def test_add_user_no_passwd(smb_conf):
    requires_passdb_modules()
    pdbl = sambacc.passdb_loader.PassDBLoader(smbconf=str(smb_conf))

    # Irritatingly, the passwd file contents appear to be cached
    # so we need to make up a user that is def. not in the etc passwd
    # equivalent, in order to get samba libs to reject it
    urec = dict(name="bob", uid=1011, gid=1011)
    ubad = sambacc.config.UserEntry(None, urec, 0)
    with pytest.raises(ValueError):
        pdbl.add_user(ubad)



#
# sambacc: a samba container configuration tool
# Copyright (C) 2021  John Mulligan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>
#

import logging
import os
import shutil

from sambacc import addc
from sambacc import samba_cmds

from .cli import best_waiter, CommandBuilder, Context, Fail

try:
    import dns
    import dns.resolver
    import dns.exception

    _DNS = True
except ImportError:
    _DNS = False


_logger = logging.getLogger(__name__)

_populated: str = "/var/lib/samba/POPULATED"
_provisioned: str = "/etc/samba/smb.conf"

dccommands = CommandBuilder()


@dccommands.command(name="summary")
def summary(ctx: Context) -> None:
    print("Hello", ctx)


_setup_choices = ["init-all", "provision", "populate", "wait-domain", "join"]


def _dosetup(ctx: Context, step_name: str) -> bool:
    setup = ctx.cli.setup or []
    return ("init-all" in setup) or (step_name in setup)


def _run_container_args(parser):
    parser.add_argument(
        "--setup",
        action="append",
        choices=_setup_choices,
        help=(
            "Specify one or more setup step names to preconfigure the"
            " container environment before the server process is started."
            " The special 'init-all' name will perform all known setup steps."
        ),
    )
    parser.add_argument(
        "--name",
        help="Specify a custom name for the dc, overriding the config file.",
    )


def _prep_provision(ctx: Context) -> None:
    if os.path.exists(_provisioned):
        _logger.info("Domain already provisioned")
        return
    domconfig = ctx.instance_config.domain()
    _logger.info(f"Provisioning domain: {domconfig.realm}")

    dcname = ctx.cli.name or domconfig.dcname
    addc.provision(
        realm=domconfig.realm,
        domain=domconfig.short_domain,
        dcname=dcname,
        admin_password=domconfig.admin_password,
    )


def _prep_join(ctx: Context) -> None:
    if os.path.exists(_provisioned):
        _logger.info("Already configured. Not joining")
        return
    domconfig = ctx.instance_config.domain()
    _logger.info(f"Provisioning domain: {domconfig.realm}")

    dcname = ctx.cli.name or domconfig.dcname
    addc.join(
        realm=domconfig.realm,
        domain=domconfig.short_domain,
        dcname=dcname,
        admin_password=domconfig.admin_password,
    )


def _prep_wait_on_domain(ctx: Context) -> None:
    if not _DNS:
        _logger.info("Can not query domain. Exiting.")
        raise Fail("no dns support available (missing dnsypthon)")

    realm = ctx.instance_config.domain().realm
    waiter = best_waiter(max_timeout=30)
    while True:
        _logger.info(f"checking for AD domain in dns: {realm}")
        try:
            dns.resolver.query(f"_ldap._tcp.{realm}.", "SRV")
            return
        except dns.exception.DNSException:
            _logger.info(f"dns record for {realm} not found")
            waiter.wait()


def _prep_populate(ctx: Context) -> None:
    if os.path.exists(_populated):
        _logger.info("populated marker exists")
        return
    _logger.info("Populating domain with default entries")

    for dgroup in ctx.instance_config.domain_groups():
        addc.create_group(dgroup.groupname)

    for duser in ctx.instance_config.domain_users():
        addc.create_user(
            name=duser.username,
            password=duser.plaintext_passwd,
            surname=duser.surname,
            given_name=duser.given_name,
        )
        # TODO: probably should improve this to avoid extra calls / loops
        for gname in duser.member_of:
            addc.add_group_members(group_name=gname, members=[duser.username])

    # "touch" the populated marker
    with open(_populated, "w"):
        pass


def _prep_krb5_conf(ctx: Context) -> None:
    shutil.copy("/var/lib/samba/private/krb5.conf", "/etc/krb5.conf")


@dccommands.command(name="run", arg_func=_run_container_args)
def run(ctx: Context) -> None:
    _logger.info("Running AD DC container")
    if _dosetup(ctx, "wait-domain"):
        _prep_wait_on_domain(ctx)
    if _dosetup(ctx, "join"):
        _prep_join(ctx)
    if _dosetup(ctx, "provision"):
        _prep_provision(ctx)
    if _dosetup(ctx, "populate"):
        _prep_populate(ctx)

    _prep_krb5_conf(ctx)
    _logger.info("Starting samba server")
    samba_cmds.execute(samba_cmds.samba_dc_foreground())



#
# sambacc: a samba container configuration tool
# Copyright (C) 2021  John Mulligan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>
#

import argparse
import functools
import logging
import typing

from sambacc import container_dns

from .cli import commands, Context, best_waiter, best_leader_locator, Fail

_logger = logging.getLogger(__name__)


def _dns_register_args(parser: argparse.ArgumentParser) -> None:
    parser.add_argument(
        "--watch",
        action="store_true",
        help=("If set, watch the source for changes and update DNS."),
    )
    parser.add_argument(
        "--domain",
        default="",
        help=("Manually specify parent domain for DNS entries."),
    )
    parser.add_argument(
        "--target",
        default=container_dns.EXTERNAL,
        choices=[container_dns.EXTERNAL, container_dns.INTERNAL],
        help=("Register IPs that fulfill the given access target."),
    )
    parser.add_argument("source", help="Path to source JSON file.")


@commands.command(name="dns-register", arg_func=_dns_register_args)
def dns_register(ctx: Context) -> None:
    """Register container & container orchestration IPs with AD DNS."""
    # This command assumes a cooperating JSON state file.
    # This file is expected to be supplied & kept up to date by
    # a container-orchestration specific component.
    iconfig = ctx.instance_config
    domain = ctx.cli.domain or ""
    if not domain:
        try:
            domain = dict(iconfig.global_options())["realm"].lower()
        except KeyError:
            raise Fail("instance not configured with domain (realm)")

    update_func = functools.partial(
        container_dns.parse_and_update,
        target_name=ctx.cli.target,
    )

    if iconfig.with_ctdb:
        _logger.info("enabling ctdb support: will check for leadership")
        update_func = _exec_if_leader(iconfig, update_func)

    if ctx.cli.watch:
        _logger.info("will watch source")
        waiter = best_waiter(ctx.cli.source)
        container_dns.watch(
            domain,
            ctx.cli.source,
            update_func,
            waiter.wait,
            print_func=print,
        )
    else:
        update_func(domain, ctx.cli.source)
    return


def _exec_if_leader(iconfig, update_func):
    def leader_update_func(
        domain: str,
        source: str,
        previous: typing.Optional[container_dns.HostState] = None,
    ) -> typing.Tuple[typing.Optional[container_dns.HostState], bool]:
        with best_leader_locator(iconfig) as ll:
            if not ll.is_leader():
                _logger.info("skipping dns update. node not leader")
                return previous, False
            _logger.info("checking for update. node is leader")
            result = update_func(domain, source, previous)
        return result

    return leader_update_func



#!/usr/bin/python3
"""Convert or compare schema files written in YAML to the corresponding
files stored in JSON.
"""

import argparse
import collections
import json
import os
import pprint
import subprocess
import sys

import yaml


nameparts = collections.namedtuple("nameparts", "full head ext")
filepair = collections.namedtuple("filepair", "origin dest format")


def _namesplit(name):
    head, ext = name.split(".", 1)
    return nameparts(name, head, ext)


def _pyname(np):
    head = np.head.replace("-", "_")
    if np.ext.startswith("schema."):
        head += "_schema"
    ext = "py"
    return nameparts(f"{head}.{ext}", head, ext)


def _format_black(path):
    # black does not formally have an api. Safeset way to use it is via
    # the cli
    path = os.path.abspath(path)
    # the --preview arg allows black to break up long strings that
    # the general check would discover and complain about. Otherwise
    # we'd be forced to ignore the formatting on these .py files.
    subprocess.run(["black", "-l78", "--preview", path], check=True)


def match(files):
    yamls = []
    jsons = []
    pys = []
    for fname in files:
        try:
            np = _namesplit(fname)
        except ValueError:
            continue
        if np.ext == "schema.yaml":
            yamls.append(np)
        if np.ext == "schema.json":
            jsons.append(np)
        if np.ext == "py":
            pys.append(np)
    pairs = []
    for yname in yamls:
        for jname in jsons:
            if jname.head == yname.head:
                pairs.append(filepair(yname, jname, "JSON"))
                break
        else:
            pairs.append(filepair(yname, None, "JSON"))
        for pyname in pys:
            if _pyname(yname).head == pyname.head:
                pairs.append(filepair(yname, pyname, "PYTHON"))
                break
        else:
            pairs.append(filepair(yname, None, "PYTHON"))
    return pairs


def report(func, path, yaml_file, json_file, fmt):
    needs_update = func(path, yaml_file, json_file, fmt)
    json_name = "---" if not json_file else json_file.full
    if not needs_update:
        print(f"{yaml_file.full} -> {fmt.lower()}:{json_name}   OK")
        return None
    print(f"{yaml_file.full} -> {fmt.lower()}:{json_name}   MISMATCH")
    return True


def update_json(path, yaml_file, json_file):
    yaml_path = os.path.join(path, yaml_file.full)
    json_path = os.path.join(path, f"{yaml_file.head}.schema.json")
    with open(yaml_path) as fh:
        yaml_data = yaml.safe_load(fh)
    with open(json_path, "w") as fh:
        json.dump(yaml_data, fh, indent=2)


def compare_json(path, yaml_file, json_file):
    if json_file is None:
        return True
    yaml_path = os.path.join(path, yaml_file.full)
    json_path = os.path.join(path, json_file.full)
    with open(yaml_path) as fh:
        yaml_data = yaml.safe_load(fh)
    with open(json_path) as fh:
        json_data = json.load(fh)
    return yaml_data != json_data


def update_py(path, yaml_file, py_file):
    yaml_path = os.path.join(path, yaml_file.full)
    py_path = os.path.join(path, _pyname(yaml_file).full)
    with open(yaml_path) as fh:
        yaml_data = yaml.safe_load(fh)
    out = []
    out.append("#!/usr/bin/python3")
    out.append("# --- GENERATED FILE --- DO NOT EDIT --- #")
    out.append(f"# --- generated from: {yaml_file.full}")
    out.append("")
    out.append(
        "SCHEMA = " + pprint.pformat(yaml_data, width=800, sort_dicts=False)
    )
    content = "\n".join(out)
    with open(py_path, "w") as fh:
        fh.write(content)
    _format_black(py_path)


def compare_py(path, yaml_file, py_file):
    if py_file is None:
        return True
    yaml_path = os.path.join(path, yaml_file.full)
    py_path = os.path.join(path, py_file.full)
    with open(yaml_path) as fh:
        yaml_data = yaml.safe_load(fh)
    with open(py_path) as fh:
        py_locals = {}
        exec(fh.read(), None, py_locals)
    py_data = py_locals.get("SCHEMA") or {}
    return yaml_data != py_data


def update(path, yaml_data, other_file, fmt):
    if fmt == "PYTHON":
        return update_py(path, yaml_data, other_file)
    return update_json(path, yaml_data, other_file)


def compare(path, yaml_data, other_file, fmt):
    if fmt == "PYTHON":
        return compare_py(path, yaml_data, other_file)
    return compare_json(path, yaml_data, other_file)


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("DIR", default=os.path.dirname(__file__), nargs="?")
    parser.add_argument("--update", action="store_true")
    cli = parser.parse_args()

    mismatches = []
    os.chdir(cli.DIR)
    fn = update if cli.update else compare
    pairs = match(os.listdir("."))
    for pair in pairs:
        mismatches.append(report(fn, ".", *pair))
    if any(mismatches):
        sys.exit(1)


if __name__ == "__main__":
    main()



#
# sambacc: a samba container configuration tool
# Copyright (C) 2021  John Mulligan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>
#

import enum
import logging
import os
import subprocess
import typing

from sambacc import config
from sambacc import jfile
from sambacc import samba_cmds
from sambacc.netcmd_loader import template_config
from sambacc.typelets import ExcType, ExcValue, ExcTraceback

_logger = logging.getLogger(__name__)


DB_DIR = "/var/lib/ctdb/persistent"
ETC_DIR = "/etc/ctdb"
SHARE_DIR = "/usr/share/ctdb"

CTDB_CONF: str = "/etc/ctdb/ctdb.conf"
CTDB_NODES: str = "/etc/ctdb/nodes"


class NodeState(str, enum.Enum):
    NEW = "new"
    READY = "ready"
    CHANGED = "changed"
    REPLACED = "replaced"
    GONE = "gone"  # reserved


def next_state(state: NodeState) -> NodeState:
    if state == NodeState.NEW:
        return NodeState.READY
    elif state == NodeState.CHANGED:
        return NodeState.REPLACED
    elif state == NodeState.REPLACED:
        return NodeState.READY
    return state


class NodeNotPresent(KeyError):
    def __init__(
        self,
        identity: typing.Any,
        pnn: typing.Optional[typing.Union[str, int]] = None,
    ) -> None:
        super().__init__(identity)
        self.identity = identity
        self.pnn = pnn


def ensure_smb_conf(
    iconfig: config.InstanceConfig, path: str = config.SMB_CONF
) -> None:
    """Ensure that the smb.conf on disk is ctdb and registry enabled."""
    with open(path, "w") as fh:
        write_smb_conf(fh, iconfig)


def write_smb_conf(fh: typing.IO, iconfig: config.InstanceConfig) -> None:
    """Write an smb.conf style output enabling ctdb and samba registry."""
    template_config(fh, iconfig.ctdb_smb_config())


def ensure_ctdb_conf(
    iconfig: config.InstanceConfig, path: str = CTDB_CONF
) -> None:
    """Ensure that the ctdb.conf on disk matches our desired state."""
    with open(path, "w") as fh:
        write_ctdb_conf(fh, iconfig.ctdb_config())


def write_ctdb_conf(
    fh: typing.IO, ctdb_params: dict, enc: typing.Callable = str
) -> None:
    """Write a ctdb.conf style output."""

    def _write_param(fh: typing.IO, name: str, key: str) -> None:
        value = ctdb_params.get(key)
        if value is None:
            return
        fh.write(enc(f"{name} = {value}\n"))

    fh.write(enc("[logging]\n"))
    _write_param(fh, "log level", "log_level")
    fh.write(enc("\n"))
    fh.write(enc("[cluster]\n"))
    _write_param(fh, "recovery lock", "recovery_lock")
    fh.write(enc("\n"))
    fh.write(enc("[legacy]\n"))
    _write_param(fh, "realtime scheduling", "realtime_scheduling")
    _write_param(fh, "script log level", "script_log_level")
    fh.write(enc("\n"))


def ensure_ctdb_nodes(
    ctdb_nodes: list[str], real_path: str, canon_path: str = CTDB_NODES
) -> None:
    """Ensure a real nodes file exists, containing the specificed content,
    and has a symlink in the proper place for ctdb.
    """
    try:
        os.unlink(canon_path)
    except FileNotFoundError:
        pass
    os.symlink(real_path, canon_path)
    # XXX: add locking?
    with open(real_path, "w") as fh:
        write_nodes_file(fh, ctdb_nodes)


def write_nodes_file(
    fh: typing.IO, ctdb_nodes: list[str], enc: typing.Callable = str
) -> None:
    """Write the ctdb nodes file."""
    for node in ctdb_nodes:
        fh.write(enc(f"{node}\n"))


def read_nodes_file(fh: typing.IO) -> list[str]:
    """Read content from an open ctdb nodes file."""
    entries = []
    for line in fh:
        entries.append(line.strip())
    return entries


def read_ctdb_nodes(path: str = CTDB_NODES) -> list[str]:
    """Read the content of the ctdb nodes file."""
    try:
        with open(path, "r") as fh:
            entries = read_nodes_file(fh)
    except FileNotFoundError:
        return []
    return entries


def ensure_ctdb_node_present(
    node: str,
    real_path: str,
    canon_path: str = CTDB_NODES,
    expected_pnn: typing.Optional[int] = None,
) -> None:
    """Ensure that the ctdb nodes file is populated with at least the
    node given. The optional `expect_pnn` can be provided to ensure that
    the node occupies the correct position in the nodes file.
    """
    nodes = read_ctdb_nodes(real_path)
    if node not in nodes:
        nodes.append(node)
    if expected_pnn is not None:
        try:
            found_pnn = nodes.index(node)
        except ValueError:
            found_pnn = -1
        if expected_pnn != found_pnn:
            raise ValueError(f"expected pnn {expected_pnn} is not {found_pnn}")
    ensure_ctdb_nodes(nodes, real_path=real_path, canon_path=canon_path)


def add_node_to_statefile(
    identity: str, node: str, pnn: int, path: str, in_nodes: bool = False
) -> None:
    """Add the given node's identity, (node) IP, and PNN to the JSON based
    state file, located at `path`. If in_nodes is true, the state file will
    reflect that the node is already added to the CTDB nodes file.
    """
    with jfile.open(path, jfile.OPEN_RW) as fh:
        jfile.flock(fh)
        data = jfile.load(fh, {})
        _update_statefile(data, identity, node, pnn, in_nodes=in_nodes)
        jfile.dump(data, fh)


def refresh_node_in_statefile(
    identity: str, node: str, pnn: int, path: str
) -> None:
    """Assuming the node is already in the statefile, update the state in
    the case that the node (IP) has changed.
    """
    with jfile.open(path, jfile.OPEN_RW) as fh:
        jfile.flock(fh)
        data = jfile.load(fh, {})
        _refresh_statefile(data, identity, node, pnn)
        jfile.dump(data, fh)


def _update_statefile(
    data: dict[str, typing.Any],
    identity: str,
    node: str,
    pnn: int,
    in_nodes: bool = False,
) -> None:
    data.setdefault("nodes", [])
    for entry in data["nodes"]:
        if pnn == entry["pnn"]:
            raise ValueError("duplicate pnn")
        if identity == entry["identity"]:
            raise ValueError("duplicate identity")
    state = NodeState.NEW
    if in_nodes:
        state = NodeState.READY
    data["nodes"].append(
        {
            "identity": identity,
            "node": node,
            "pnn": pnn,
            "state": state,
        }
    )


def _refresh_statefile(
    data: dict[str, typing.Any],
    identity: str,
    node: str,
    pnn: int,
    in_nodes: bool = False,
) -> None:
    data.setdefault("nodes", [])
    node_entry = None
    for entry in data["nodes"]:
        if pnn == entry["pnn"] and identity == entry["identity"]:
            node_entry = entry
            break
        if pnn == entry["pnn"]:
            raise ValueError(
                f"matching pnn ({pnn}) identity={entry['identity']}"
            )
    if not node_entry:
        raise NodeNotPresent(identity, pnn)
    if node_entry["node"] == node:
        # do nothing
        return
    node_entry["node"] = node
    node_entry["state"] = NodeState.CHANGED


def _get_state(entry: dict[str, typing.Any]) -> NodeState:
    return NodeState(entry["state"])


def _get_state_ok(entry: dict[str, typing.Any]) -> bool:
    return _get_state(entry) == NodeState.READY


def pnn_in_nodes(pnn: int, nodes_json: str, real_path: str) -> bool:
    """Returns true if the specified pnn has an entry in the nodes json
    file and that the node is already added to the ctdb nodes file.
    """
    with jfile.open(nodes_json, jfile.OPEN_RO) as fh:
        jfile.flock(fh)
        json_data = jfile.load(fh, {})
        current_nodes = json_data.get("nodes", [])
        for entry in current_nodes:
            if pnn == entry["pnn"] and _get_state_ok(entry):
                return True
    return False


def manage_nodes(
    pnn: int,
    nodes_json: str,
    real_path: str,
    pause_func: typing.Callable,
) -> None:
    """Monitor nodes json for updates, reflecting those changes into ctdb."""
    while True:
        _logger.info("checking if node is able to make updates")
        if _node_check(pnn, nodes_json, real_path):
            _logger.info("checking for node updates")
            if _node_update(nodes_json, real_path):
                _logger.info("updated nodes")
        else:
            _logger.warning("node can not make updates")
        pause_func()


def _node_check(pnn: int, nodes_json: str, real_path: str) -> bool:
    with jfile.open(nodes_json, jfile.OPEN_RO) as fh:
        jfile.flock(fh)
        desired = jfile.load(fh, {}).get("nodes", [])
    ctdb_nodes = read_ctdb_nodes(real_path)
    # first: check to see if the current node is in the nodes file
    try:
        my_desired = [e for e in desired if e.get("pnn") == pnn][0]
    except IndexError:
        # no entry found for this node
        _logger.warning(f"PNN {pnn} not found in json state file")
        return False
    if my_desired["node"] not in ctdb_nodes:
        # this current node is not in the nodes file.
        # it is ineligible to make changes to the nodes file
        return False
    # this node is already in the nodes file!
    return True


def _node_update_check(
    json_data: dict[str, typing.Any], nodes_json: str, real_path: str
) -> tuple[list[str], list[typing.Any], list[typing.Any]]:
    desired = json_data.get("nodes", [])
    ctdb_nodes = read_ctdb_nodes(real_path)
    update_nodes = []
    need_reload = []
    _update_states = (NodeState.NEW, NodeState.CHANGED, NodeState.REPLACED)
    for entry in desired:
        pnn = entry["pnn"]
        matched = _node_line(ctdb_nodes, pnn) == entry["node"]
        if matched and _get_state_ok(entry):
            # everything's fine. skip this entry
            continue
        elif not matched:
            if entry["state"] in _update_states:
                update_nodes.append(entry)
                need_reload.append(entry)
            elif entry["state"] == NodeState.READY:
                msg = f"ready node (pnn {pnn}) missing from {ctdb_nodes}"
                raise ValueError(msg)
        else:
            # node present but state indicates
            # update is not finalized
            need_reload.append(entry)
    return ctdb_nodes, update_nodes, need_reload


def _node_line(ctdb_nodes: list[str], pnn: int) -> str:
    try:
        return ctdb_nodes[pnn]
    except IndexError:
        return ""


def _entry_to_node(ctdb_nodes: list[str], entry: dict[str, typing.Any]) -> str:
    pnn: int = entry["pnn"]
    if entry["state"] == NodeState.CHANGED:
        return "#{}".format(ctdb_nodes[pnn].strip("#"))
    return entry["node"]


def _node_update(nodes_json: str, real_path: str) -> bool:
    # open r/o so that we don't initailly open for write.  we do a probe and
    # decide if anything needs to be updated if we are wrong, its not a
    # problem, we'll "time out" and reprobe later
    with jfile.open(nodes_json, jfile.OPEN_RO) as fh:
        jfile.flock(fh)
        json_data = jfile.load(fh, {})
        _, test_chg_nodes, test_need_reload = _node_update_check(
            json_data, nodes_json, real_path
        )
        if not test_chg_nodes and not test_need_reload:
            _logger.info("examined nodes state - no changes")
            return False
    # we probably need to make a change. but we recheck our state again
    # under lock, with the data file open r/w
    # update the nodes file and make changes to ctdb
    with jfile.open(nodes_json, jfile.OPEN_RW) as fh:
        jfile.flock(fh)
        json_data = jfile.load(fh, {})
        ctdb_nodes, chg_nodes, need_reload = _node_update_check(
            json_data, nodes_json, real_path
        )
        if not chg_nodes and not need_reload:
            _logger.info("reexamined nodes state - no changes")
            return False
        _logger.info("writing updates to ctdb nodes file")
        new_ctdb_nodes = list(ctdb_nodes)
        for entry in chg_nodes:
            pnn = entry["pnn"]
            expected_line = _entry_to_node(ctdb_nodes, entry)
            if _node_line(new_ctdb_nodes, pnn) == expected_line:
                continue
            if entry["state"] == NodeState.NEW:
                if pnn != len(new_ctdb_nodes):
                    raise ValueError(
                        f"unexpected pnn in new entry {entry}:"
                        " nodes: {new_ctdb_nodes}"
                    )
                new_ctdb_nodes.append(expected_line)
            else:
                new_ctdb_nodes[pnn] = expected_line
        with open(real_path, "w") as nffh:
            write_nodes_file(nffh, new_ctdb_nodes)
            nffh.flush()
            os.fsync(nffh)
        _logger.info("running: ctdb reloadnodes")
        subprocess.check_call(list(samba_cmds.ctdb["reloadnodes"]))
        for entry in need_reload:
            entry["state"] = next_state(entry["state"])
            _logger.debug(
                "setting node identity=[{}] pnn={} to {}".format(
                    entry["identity"],
                    entry["pnn"],
                    entry["state"],
                )
            )
        jfile.dump(json_data, fh)
        fh.flush()
        os.fsync(fh)
    return True


def ensure_ctdbd_etc_files(
    etc_path: str = ETC_DIR, src_path: str = SHARE_DIR
) -> None:
    """Ensure certain files that ctdbd expects to exist in its etc dir
    do exist.
    """
    functions_src = os.path.join(src_path, "functions")
    functions_dst = os.path.join(etc_path, "functions")
    notify_src = os.path.join(src_path, "notify.sh")
    notify_dst = os.path.join(etc_path, "notify.sh")
    legacy_scripts_src = os.path.join(src_path, "events/legacy")
    legacy_scripts_dst = os.path.join(etc_path, "events/legacy")
    link_legacy_scripts = ["00.ctdb.script"]

    os.makedirs(etc_path, exist_ok=True)
    try:
        os.unlink(functions_dst)
    except FileNotFoundError:
        pass
    os.symlink(functions_src, functions_dst)

    try:
        os.unlink(notify_dst)
    except FileNotFoundError:
        pass
    os.symlink(notify_src, notify_dst)

    os.makedirs(legacy_scripts_dst, exist_ok=True)
    for legacy_script_name in link_legacy_scripts:
        lscript_src = os.path.join(legacy_scripts_src, legacy_script_name)
        lscript_dst = os.path.join(legacy_scripts_dst, legacy_script_name)
        try:
            os.unlink(lscript_dst)
        except FileNotFoundError:
            pass
        os.symlink(lscript_src, lscript_dst)


_SRC_TDB_FILES = [
    "account_policy.tdb",
    "group_mapping.tdb",
    "passdb.tdb",
    "registry.tdb",
    "secrets.tdb",
    "share_info.td",
    "winbindd_idmap.tdb",
]

_SRC_TDB_DIRS = [
    "/var/lib/samba",
    "/var/lib/samba/private",
]


def migrate_tdb(
    iconfig: config.InstanceConfig, dest_dir: str, pnn: int = 0
) -> None:
    """Migrate TDB files into CTDB."""
    # TODO: these paths should be based on our instance config, not hard coded
    for tdbfile in _SRC_TDB_FILES:
        for parent in _SRC_TDB_DIRS:
            tdb_path = os.path.join(parent, tdbfile)
            if _has_tdb_file(tdb_path):
                _convert_tdb_file(tdb_path, dest_dir, pnn=pnn)


def _has_tdb_file(tdb_path: str) -> bool:
    # TODO: It would be preferable to handle errors from the convert
    # function only, but it if ltdbtool is missing it raises FileNotFoundError
    # and its not simple to disambiguate between the command missing and the
    # tdb file missing.
    _logger.info(f"Checking for {tdb_path}")
    return os.path.isfile(tdb_path)


def _convert_tdb_file(tdb_path: str, dest_dir: str, pnn: int = 0) -> None:
    orig_name = os.path.basename(tdb_path)
    opath = os.path.join(dest_dir, f"{orig_name}.{pnn}")
    _logger.info(f"Converting {tdb_path} to {opath} ...")
    cmd = samba_cmds.ltdbtool["convert", "-s0", tdb_path, opath]
    subprocess.check_call(list(cmd))


def check_nodestatus(cmd: samba_cmds.SambaCommand = samba_cmds.ctdb) -> None:
    cmd_ctdb_check = cmd["nodestatus"]
    samba_cmds.execute(cmd_ctdb_check)


class CLILeaderStatus:
    _isleader = False

    def is_leader(self) -> bool:
        return self._isleader


class CLILeaderLocator:
    """A leader locator that relies entirely on checking the
    recovery master using the ctdb command line tool.
    """

    def __enter__(self) -> CLILeaderStatus:
        mypnn = recmaster = ""
        # mypnn = <ctdb pnn>
        pnn_cmd = samba_cmds.ctdb["pnn"]
        try:
            out = subprocess.check_output(list(pnn_cmd))
            mypnn = out.decode("utf8").strip()
        except subprocess.CalledProcessError as err:
            _logger.error(f"command {pnn_cmd!r} failed: {err!r}")
        except FileNotFoundError:
            _logger.error(f"ctdb command ({pnn_cmd!r}) not found")
        # recmaster = <ctdb recmaster|leader>
        admin_cmd = samba_cmds.ctdb_leader_admin_cmd()
        recmaster_cmd = samba_cmds.ctdb[admin_cmd]
        try:
            out = subprocess.check_output(list(recmaster_cmd))
            recmaster = out.decode("utf8").strip()
        except subprocess.CalledProcessError as err:
            _logger.error(f"command {recmaster_cmd!r} failed: {err!r}")
        except FileNotFoundError:
            _logger.error(f"ctdb command ({recmaster_cmd!r}) not found")

        sts = CLILeaderStatus()
        sts._isleader = bool(mypnn) and mypnn == recmaster
        return sts

    def __exit__(
        self, exc_type: ExcType, exc_val: ExcValue, exc_tb: ExcTraceback
    ) -> bool:
        return True



#
# sambacc: a samba container configuration tool
# Copyright (C) 2021  John Mulligan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>
#

import argparse
import logging
import os
import socket
import typing

from sambacc import ctdb

from .cli import best_waiter, commands, Context, Fail

_logger = logging.getLogger(__name__)

# Rather irritatingly, k8s does not have a simple method for passing the
# ordinal index of a stateful set down to the containers. This has been
# proposed but not implemented yet. See:
#  https://github.com/kubernetes/kubernetes/issues/40651
# While I find putting any k8s specific knowledge in sambacc distasteful
# all we're really doing is teaching sambacc how to extract the node
# number from the host name, an operation that's not k8s specific.
# That isn't *too* dirty. Just a smudge really. :-)
_AFTER_LAST_DASH = "after-last-dash"


def _ctdb_ok():
    sambacc_ctdb = os.environ.get("SAMBACC_CTDB")
    gate = "ctdb-is-experimental"
    if sambacc_ctdb == gate:
        return
    print("Using CTDB with samba-container (sambacc) is experimental.")
    print("If you are developing or testing features for sambacc please")
    print("set the environment variable SAMBACC_CTDB to the value:")
    print("    ", gate)
    print("before continuing and try again.")
    print()
    raise Fail(gate)


def _ctdb_migrate_args(parser: argparse.ArgumentParser) -> None:
    parser.add_argument(
        "--dest-dir",
        default=ctdb.DB_DIR,
        help=("Specify where CTDB database files will be written."),
    )


def _ctdb_general_node_args(parser: argparse.ArgumentParser) -> None:
    parser.add_argument(
        "--hostname",
        help=("Specify the host name for the CTDB node"),
    )
    parser.add_argument(
        "--node-number",
        type=int,
        help=("Expected node number"),
    )
    # This is a choice with a single acceptable param, rather than an on/off
    # bool, # in the case that other container orchs have a similar but not
    # quite the same issue and we want to support a different scheme someday.
    parser.add_argument(
        "--take-node-number-from-hostname",
        choices=(_AFTER_LAST_DASH,),
        help=(
            "Take the node number from the given host name following"
            " the specified policy."
        ),
    )
    parser.add_argument(
        "--persistent-path",
        help=("Path to a persistent path for storing nodes file"),
    )


def _ctdb_set_node_args(parser: argparse.ArgumentParser) -> None:
    _ctdb_general_node_args(parser)
    parser.add_argument(
        "--ip",
        help=("Specify node by IP"),
    )


class NodeParams:
    _ctx: Context
    node_number: typing.Optional[int] = None
    hostname: typing.Optional[str] = None
    persistent_path: str = ""
    nodes_json: str = ""
    _ip_addr: typing.Optional[str] = None

    def __init__(self, ctx: Context):
        self._ctx = ctx
        ccfg = ctx.instance_config.ctdb_config()

        # stuff that many of the commands use
        self.persistent_path = ctx.cli.persistent_path
        if self.persistent_path is None:
            self.persistent_path = ccfg["nodes_path"]
        self.nodes_json = ccfg["nodes_json"]

        self.hostname = ctx.cli.hostname
        if ctx.cli.node_number is not None:
            if ctx.cli.node_number < 0:
                raise ValueError(f"invalid node number: {ctx.cli.node_number}")
            self.node_number = ctx.cli.node_number
        elif ctx.cli.take_node_number_from_hostname == _AFTER_LAST_DASH:
            if not self.hostname:
                raise ValueError(
                    "--hostname required if taking node number from host name"
                )
            if "-" not in self.hostname:
                raise ValueError(
                    f"invalid hostname for node number: {self.hostname}"
                )
            self.node_number = int(self.hostname.rsplit("-")[-1])
        else:
            self.node_number = None

    @property
    def node_ip_addr(self) -> str:
        if self._ip_addr is None:
            cli = self._ctx.cli
            if getattr(cli, "ip", None):
                self._ip_addr = cli.ip
            elif cli.hostname:
                self._ip_addr = _lookup_hostname(cli.hostname)
            else:
                raise ValueError("can not determine node ip")
        return self._ip_addr

    @property
    def identity(self):
        # this could be extended to use something like /etc/machine-id
        # or whatever in the future.
        if self.hostname:
            return self.hostname
        elif self.node_number:
            return f"node-{self.node_number}"
        else:
            # the dashes make this an invalid dns name
            return "-unknown-"


@commands.command(name="ctdb-migrate", arg_func=_ctdb_migrate_args)
def ctdb_migrate(ctx: Context) -> None:
    """Migrate standard samba databases to CTDB databases."""
    _ctdb_ok()
    ctdb.migrate_tdb(ctx.instance_config, ctx.cli.dest_dir)


def _lookup_hostname(hostname):
    # XXX this is a nasty little hack.
    ips = socket.gethostbyname_ex(hostname)[2]
    addr = [ip for ip in ips if ip != "127.0.0.1"][0]
    _logger.info(f"Determined address for {hostname}: {addr}")
    return addr


@commands.command(name="ctdb-set-node", arg_func=_ctdb_set_node_args)
def ctdb_set_node(ctx: Context) -> None:
    """Set up the current node in the ctdb and sambacc nodes files."""
    _ctdb_ok()
    np = NodeParams(ctx)
    expected_pnn = np.node_number

    try:
        ctdb.refresh_node_in_statefile(
            np.identity,
            np.node_ip_addr,
            int(expected_pnn or 0),
            path=np.nodes_json,
        )
        return
    except ctdb.NodeNotPresent:
        pass

    ctdb.add_node_to_statefile(
        np.identity,
        np.node_ip_addr,
        int(expected_pnn or 0),
        path=np.nodes_json,
        in_nodes=(expected_pnn == 0),
    )
    if expected_pnn == 0:
        ctdb.ensure_ctdb_node_present(
            node=np.node_ip_addr,
            expected_pnn=expected_pnn,
            real_path=np.persistent_path,
        )


@commands.command(name="ctdb-manage-nodes", arg_func=_ctdb_general_node_args)
def ctdb_manage_nodes(ctx: Context) -> None:
    """Run a long lived procees to monitor the node state file for new nodes.
    When a new node is found, if the current node is in the correct state, this
    node will add it to CTDB.
    """
    _ctdb_ok()
    np = NodeParams(ctx)
    expected_pnn = np.node_number or 0
    waiter = best_waiter(np.nodes_json)

    errors = 0
    while True:
        try:
            ctdb.manage_nodes(
                expected_pnn,
                nodes_json=np.nodes_json,
                real_path=np.persistent_path,
                pause_func=waiter.wait,
            )
            errors = 0
        except KeyboardInterrupt:
            raise
        except Exception as err:
            _logger.error(
                f"error during manage_nodes: {err}, count={errors}",
                exc_info=True,
            )
            errors += 1
            if errors > 10:
                _logger.error(f"too many retries ({errors}). giving up")
                raise
            waiter.wait()


@commands.command(name="ctdb-must-have-node", arg_func=_ctdb_general_node_args)
def ctdb_must_have_node(ctx: Context) -> None:
    """Block until the current node is present in the ctdb nodes file."""
    _ctdb_ok()
    np = NodeParams(ctx)
    expected_pnn = np.node_number or 0
    waiter = best_waiter(np.nodes_json)

    while True:
        if ctdb.pnn_in_nodes(
            expected_pnn,
            nodes_json=np.nodes_json,
            real_path=np.persistent_path,
        ):
            return
        _logger.info("node not yet ready")
        waiter.wait()



#
# sambacc: a samba container configuration tool
# Copyright (C) 2021  John Mulligan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>
#

import argparse
import functools
import logging
import subprocess
import sys
import typing

from sambacc import config
from sambacc import samba_cmds
from sambacc.simple_waiter import watch
import sambacc.netcmd_loader as nc
import sambacc.paths as paths

from .cli import (
    Context,
    best_leader_locator,
    best_waiter,
    commands,
    perms_handler,
    setup_steps,
)

_logger = logging.getLogger(__name__)


@commands.command(name="print-config")
def print_config(ctx: Context) -> None:
    """Display the samba configuration sourced from the sambacc config
    in the format of smb.conf.
    """
    nc.template_config(sys.stdout, ctx.instance_config)


@commands.command(name="import")
@setup_steps.command(name="config")
def import_config(ctx: Context) -> None:
    """Import configuration parameters from the sambacc config to
    samba's registry config.
    """
    # there are some expectations about what dirs exist and perms
    paths.ensure_samba_dirs()

    loader = nc.NetCmdLoader()
    loader.import_config(ctx.instance_config)


def _update_config_args(parser: argparse.ArgumentParser) -> None:
    parser.add_argument(
        "--watch",
        action="store_true",
        help=("If set, watch the source for changes and update config."),
    )


def _read_config(ctx: Context) -> config.InstanceConfig:
    cfgs = ctx.cli.config or []
    return config.read_config_files(
        cfgs, require_validation=ctx.require_validation
    ).get(ctx.cli.identity)


def _update_config(
    current: config.InstanceConfig,
    previous: typing.Optional[config.InstanceConfig],
    ensure_paths: bool = True,
    notify_server: bool = True,
) -> typing.Tuple[config.InstanceConfig, bool]:
    """Compare the current and previous instance configurations. If they
    differ, ensure any new paths, update the samba config, and inform any
    running smbds of the new configuration.  Return the current config and a
    boolean indicating if the instance configs differed.
    """
    # has the config changed?
    changed = current != previous
    # ensure share paths exist
    if changed and ensure_paths:
        for share in current.shares():
            path = share.path()
            if not path:
                continue
            _logger.info(f"Ensuring share path: {path}")
            paths.ensure_share_dirs(path)
            _logger.info(f"Updating permissions if needed: {path}")
            perms_handler(share.permissions_config(), path).update()
    # update smb config
    if changed:
        _logger.info("Updating samba configuration")
        loader = nc.NetCmdLoader()
        loader.import_config(current)
    # notify smbd of changes
    if changed and notify_server:
        subprocess.check_call(
            list(samba_cmds.smbcontrol["smbd", "reload-config"])
        )
    return current, changed


def _exec_if_leader(
    ctx: Context,
    cond_func: typing.Callable[..., typing.Tuple[config.InstanceConfig, bool]],
) -> typing.Callable[..., typing.Tuple[config.InstanceConfig, bool]]:
    """Run the cond func only on "nodes" that are the cluster leader."""

    # CTDB status and leader detection is not changeable at runtime.
    # we do not need to account for it changing in the updated config file(s)
    @functools.wraps(cond_func)
    def _call_if_leader(
        current: config.InstanceConfig, previous: config.InstanceConfig
    ) -> typing.Tuple[config.InstanceConfig, bool]:
        with best_leader_locator(ctx.instance_config) as ll:
            if not ll.is_leader():
                _logger.info("skipping config update. node not leader")
                return previous, False
            _logger.info("checking for update. node is leader")
            result = cond_func(current, previous)
        return result

    return _call_if_leader


@commands.command(name="update-config", arg_func=_update_config_args)
def update_config(ctx: Context) -> None:
    _get_config = functools.partial(_read_config, ctx)
    _cmp_func = _update_config

    if ctx.instance_config.with_ctdb:
        _logger.info("enabling ctdb support: will check for leadership")
        _cmp_func = _exec_if_leader(ctx, _cmp_func)

    if ctx.cli.watch:
        _logger.info("will watch configuration source")
        waiter = best_waiter(ctx.cli.config)
        watch(
            waiter,
            ctx.instance_config,
            _get_config,
            _cmp_func,
        )
    else:
        # we pass None as the previous config so that the command is
        # not nearly always a no-op when run from the command line.
        _cmp_func(_get_config(), None)
    return



#
# sambacc: a samba container configuration tool
# Copyright (C) 2021  John Mulligan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>
#

import argparse
import logging
import os
import time
import typing

from sambacc import config
from sambacc import samba_cmds

from . import check  # noqa: F401
from . import config as config_cmds
from . import ctdb  # noqa: F401
from . import dns  # noqa: F401
from . import initialize  # noqa: F401
from . import join  # noqa: F401
from . import run  # noqa: F401
from . import users  # noqa: F401
from .cli import commands, Fail, Parser

DEFAULT_CONFIG = "/etc/samba/container/config.json"
DEFAULT_JOIN_MARKER = "/var/lib/samba/container-join-marker.json"

default_cfunc = config_cmds.print_config


def global_args(parser: Parser) -> None:
    parser.add_argument(
        "--config",
        action="append",
        help=(
            "Specify source configuration"
            " (can also be set in the environment by SAMBACC_CONFIG)."
        ),
    )
    parser.add_argument(
        "--identity",
        help=(
            "A string identifying the local identity"
            " (can also be set in the environment by SAMBA_CONTAINER_ID)."
        ),
    )
    parser.add_argument(
        "--etc-passwd-path",
        default="/etc/passwd",
        help="Specify a path for the passwd file.",
    )
    parser.add_argument(
        "--etc-group-path",
        default="/etc/group",
        help="Specify a path for the group file.",
    )
    parser.add_argument(
        "--username",
        default="Administrator",
        help="Specify a user name for domain access.",
    )
    parser.add_argument(
        "--password", default="", help="Specify a password for domain access."
    )
    parser.add_argument(
        "--debug-delay",
        type=int,
        help="Delay activity for a specified number of seconds.",
    )
    parser.add_argument(
        "--join-marker",
        default=DEFAULT_JOIN_MARKER,
        help="Path to a file used to indicate a join has been peformed.",
    )
    parser.add_argument(
        "--samba-debug-level",
        choices=[str(v) for v in range(0, 11)],
        help="Specify samba debug level for commands.",
    )
    parser.add_argument(
        "--samba-command-prefix",
        help="Wrap samba commands within a supplied command prefix",
    )
    parser.add_argument(
        "--skip-if-file",
        action="append",
        help=("Perform no action if the specified path exists."),
    )
    parser.add_argument(
        "--validate-config",
        choices=("auto", "required", "true", "false"),
        help=("Perform schema based validation of configuration."),
    )


def from_env(
    ns: typing.Any,
    var: str,
    ename: str,
    default: typing.Any = None,
    vtype: typing.Optional[typing.Callable] = str,
) -> None:
    value = getattr(ns, var, None)
    if not value:
        value = os.environ.get(ename, "")
    if vtype is not None:
        value = vtype(value)
    if value:
        setattr(ns, var, value)


def split_paths(value):
    if not value:
        return value
    if not isinstance(value, list):
        value = [value]
    out = []
    for v in value:
        for part in v.split(":"):
            out.append(part)
    return out


def env_to_cli(cli: typing.Any) -> None:
    from_env(
        cli,
        "config",
        "SAMBACC_CONFIG",
        vtype=split_paths,
        default=DEFAULT_CONFIG,
    )
    from_env(
        cli,
        "join_files",
        "SAMBACC_JOIN_FILES",
        vtype=split_paths,
    )
    from_env(cli, "identity", "SAMBA_CONTAINER_ID")
    from_env(cli, "username", "JOIN_USERNAME")
    from_env(cli, "password", "INSECURE_JOIN_PASSWORD")
    from_env(cli, "samba_debug_level", "SAMBA_DEBUG_LEVEL")
    from_env(cli, "validate_config", "SAMBACC_VALIDATE_CONFIG")


class CommandContext:
    """CLI Context for standard samba-container commands."""

    def __init__(self, cli_args: argparse.Namespace):
        self._cli = cli_args
        self._iconfig: typing.Optional[config.InstanceConfig] = None
        self.expects_ctdb = False

    @property
    def cli(self) -> argparse.Namespace:
        return self._cli

    @property
    def instance_config(self) -> config.InstanceConfig:
        if self._iconfig is None:
            cfgs = self.cli.config or []
            self._iconfig = config.read_config_files(
                cfgs, require_validation=self.require_validation
            ).get(self.cli.identity)
        return self._iconfig

    @property
    def require_validation(self) -> typing.Optional[bool]:
        if self.cli.validate_config in ("required", "true"):
            return True
        if self.cli.validate_config == "false":
            return False
        return None


def pre_action(cli: typing.Any) -> None:
    """Handle debugging/diagnostic related options before the target
    action of the command is performed.
    """
    if cli.debug_delay:
        time.sleep(int(cli.debug_delay))
    if cli.samba_debug_level:
        samba_cmds.set_global_debug(cli.samba_debug_level)
    if cli.samba_command_prefix:
        samba_cmds.set_global_prefix([cli.samba_command_prefix])


def enable_logging(cli: typing.Any) -> None:
    logger = logging.getLogger()
    logger.setLevel(logging.INFO)
    handler = logging.StreamHandler()
    handler.setFormatter(
        logging.Formatter("{asctime}: {levelname}: {message}", style="{")
    )
    handler.setLevel(logging.INFO)
    logger.addHandler(handler)


def action_filter(cli: typing.Any) -> typing.Optional[str]:
    for path in cli.skip_if_file or []:
        if os.path.exists(path):
            return f"skip-if-file: {path} exists"
    return None


def main(args: typing.Optional[typing.Sequence[str]] = None) -> None:
    cli = commands.assemble(arg_func=global_args).parse_args(args)
    env_to_cli(cli)
    enable_logging(cli)
    if not cli.identity:
        raise Fail("missing container identity")

    pre_action(cli)
    skip = action_filter(cli)
    if skip:
        print(f"Action skipped: {skip}")
        return
    cfunc = getattr(cli, "cfunc", default_cfunc)
    cfunc(CommandContext(cli))
    return


if __name__ == "__main__":
    main()



#
# sambacc: a samba container configuration tool
# Copyright (C) 2021  John Mulligan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>
#

import typing

from . import addc
from .cli import Fail
from .main import (
    CommandContext,
    action_filter,
    enable_logging,
    env_to_cli,
    global_args,
    pre_action,
)


default_cfunc = addc.summary


def main(args: typing.Optional[typing.Sequence[str]] = None) -> None:
    cli = addc.dccommands.assemble(arg_func=global_args).parse_args(args)
    env_to_cli(cli)
    enable_logging(cli)
    if not cli.identity:
        raise Fail("missing container identity")

    pre_action(cli)
    skip = action_filter(cli)
    if skip:
        print(f"Action skipped: {skip}")
        return
    cfunc = getattr(cli, "cfunc", default_cfunc)
    cfunc(CommandContext(cli))
    return


if __name__ == "__main__":
    main()
